"""Generic Intelligent Document Processing generic keywords capable of working with
various engines.

Currently, supporting the following:
- Google Document AI (`rpaframework-google`)
- Base64
- Nanonets
"""

import functools
import logging
from enum import Enum
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

from robot.api.deco import keyword, library

from RPA.JSON import JSONType
from RPA.Robocorp.Vault import Vault
from RPA.Robocorp.utils import PathType
from RPA.RobotLogListener import RobotLogListener


lib_vault = Vault()

SecretType = Optional[Union[str, Path, Tuple, List, Dict]]
ResultType = JSONType

try:
    from google.cloud import documentai_v1 as documentai
except ImportError:
    pass
else:
    ResultType = Union[ResultType, documentai.Document]


class EngineName(Enum):
    """Supported engines to process documents with."""

    GOOGLE = "google"
    BASE64 = "base64ai"
    NANONETS = "nanonets"


@library
class DocumentAI:
    """Wrapper library offering generic keywords for initializing, scanning and
    retrieving results as fields from documents (PDF, PNG etc.).

    Library requires at the minimum `rpaframework` version **19.0.0**.

    This is a helper facade for the following libraries:

    - RPA.Cloud.Google (requires `rpaframework-google`)
    - RPA.DocumentAI.Base64AI
    - RPA.DocumentAI.Nanonets

    Where the following steps are required:

    1. Engine initialization: ``Init Engine``
    2. Document scan: ``Predict``
    3. Result retrieval: ``Get Result``

    So no matter the engine you're using, the very same keywords can be used, as only
    the passed parameters will differ (please check the docs on each library for
    particularities). Once initialized, you can jump between the engines with
    ``Switch Engine``. Before scanning documents, you must configure the service first,
    with a model to scan the files with and an API key for authorizing the access.

    See Portal example: https://robocorp.com/portal/robot/robocorp/example-document-ai

    **Example: Robot Framework**

    .. code-block:: robotframework

        *** Settings ***
        Library    RPA.DocumentAI

        *** Tasks ***
        Scan Documents
            Init Engine    base64ai    vault=document_ai:base64ai
            Init Engine    nanonets    vault=document_ai:nanonets

            Switch Engine   base64ai
            Predict    invoice.png
            ${data} =    Get Result
            Log List    ${data}

            Switch Engine   nanonets
            Predict    invoice.png      model=858e4b37-6679-4552-9481-d5497dfc0b4a
            ${data} =    Get Result
            Log List    ${data}

    **Example: Python**

    .. code-block:: python

        from RPA.DocumentAI import DocumentAI, EngineName

        lib_docai = DocumentAI()
        lib_docai.init_engine(
            EngineName.GOOGLE, vault="document_ai:serviceaccount", region="eu"
        )
        lib_docai.predict(
            "invoice.pdf", model="df1d166771005ff4",
            project_id="complete-agency-347912", region="eu"
        )
        print(lib_docai.get_result())
    """

    ROBOT_LIBRARY_SCOPE = "GLOBAL"
    ROBOT_LIBRARY_DOC_FORMAT = "REST"

    def __init__(self):
        listener = RobotLogListener()
        listener.register_protected_keywords(["RPA.DocumentAI.init_engine"])
        self.logger = logging.getLogger(__name__)

        self._active_engine: Optional[EngineName] = None
        self._engines: Dict[EngineName, Any] = {}
        self._results: Dict[EngineName, ResultType] = {}

    def _check_engine(self):
        if not self._active_engine:
            raise RuntimeError(
                "can't execute without an engine set, please run"
                " `Init Engine    <name>    ...` first"
            )

    @property
    def engine(self) -> Any:
        self._check_engine()
        return self._engines[self._active_engine]

    @property
    def result(self) -> ResultType:
        self._check_engine()
        result = self._results.get(self._active_engine)
        if not result:
            raise RuntimeError(
                "there's no result obtained yet, please run"
                " `Predict    <location>    ...` first"
            )

        return result

    def _get_secret_value(
        self, secret: SecretType, vault: Optional[Dict]
    ) -> SecretType:
        if vault:
            assert (
                len(vault) == 1
            ), "`vault` should contain one key (Vault name) and one value (secret key)"
            name, key = list(vault.items())[0]
            secrets = lib_vault.get_secret(name)
            self.logger.debug("Using secret from the Vault.")
            return secrets[key]

        if not secret:
            self.logger.debug("Using secret implicitly from environment variable(s).")
            return None

        secret_path = (
            Path(secret).expanduser().resolve() if isinstance(secret, str) else None
        )
        if secret_path and secret_path.exists():
            # File-based secret, don't return its content, but its file location
            #  as object instead. (the engine itself knows how to use it from there)
            self.logger.debug("Using secret from local file path at: %s", secret_path)
            return secret_path

        self.logger.debug("Using secret as provided.")
        return secret  # secret in plain text or object

    def _init_google(
        self,
        secret_value: SecretType,
        vault: Optional[Dict] = None,
        auth_type: Optional[str] = None,
        region: Optional[str] = None,
    ):
        # pylint: disable=import-outside-toplevel
        try:
            from RPA.Cloud.Google import Google
        except ImportError as exc:
            raise ImportError(
                "dependency `rpaframework-google` needs to be installed in order to"
                " use the 'google' engine"
            ) from exc

        lib_kwargs = {}
        init_kwargs = {}
        if vault:
            # Vault is enabled, so mark the same in the engine library as well.
            name, key = list(vault.items())[0]
            lib_kwargs.update(
                {
                    "vault_name": name,
                    "vault_secret_key": key,
                }
            )
            init_kwargs["use_robocorp_vault"] = True
        elif secret_value:
            # Vault not used, therefore the provided secret is a file path pointing to
            #  a service account or token JSON file.
            if not isinstance(secret_value, Path):
                raise ValueError(
                    f"secret {secret_value!r} not supported,"
                    " a file path holding the content is needed"
                )
            if (auth_type or "serviceaccount") == "serviceaccount":
                secret_type = "service_account"
            else:
                secret_type = "token_file"
            init_kwargs[secret_type] = str(secret_value)
        if auth_type:
            lib_kwargs["cloud_auth_type"] = auth_type
        if region:
            init_kwargs["region"] = region

        engine = Google(**lib_kwargs)
        engine.init_document_ai(**init_kwargs)
        self._engines[EngineName.GOOGLE] = engine

    @staticmethod
    def _secret_value_to_params(secret_value: SecretType) -> Tuple[Tuple, Dict]:
        args, kwargs = (), {}
        if isinstance(secret_value, (tuple, list)):
            args = tuple(secret_value)
        elif isinstance(secret_value, dict):
            kwargs = secret_value
        elif isinstance(secret_value, str):
            for part in secret_value.split(","):
                part = part.strip()
                if ":" in part:
                    key, value = part.split(":", 1)
                    kwargs[key] = value
                else:
                    args += (part,)
        else:
            raise TypeError(f"not supported secret type {type(secret_value)}")

        return args, kwargs

    def _init_base64(self, secret_value: SecretType):
        # pylint: disable=import-outside-toplevel
        from RPA.DocumentAI.Base64AI import Base64AI

        engine = Base64AI()
        args, kwargs = self._secret_value_to_params(secret_value)
        engine.set_authorization(*args, **kwargs)
        self._engines[EngineName.BASE64] = engine

    def _init_nanonets(self, secret_value: SecretType):
        # pylint: disable=import-outside-toplevel
        from RPA.DocumentAI.Nanonets import Nanonets

        engine = Nanonets()
        args, kwargs = self._secret_value_to_params(secret_value)
        engine.set_authorization(*args, **kwargs)
        self._engines[EngineName.NANONETS] = engine

    @keyword
    def switch_engine(self, name: Union[EngineName, str]) -> None:
        """Switch between already initialized engines.

        Use this to jump between engines when scanning with multiple of them.

        :param name: Name of the engine to be set as active. (choose between: %s)

        **Example: Robot Framework**

        .. code-block:: robotframework

            *** Tasks ***
            Document AI All
                @{engines} =    Create List     base64ai    nanonets
                FOR    ${engine}    IN    @{engines}
                    Switch Engine    ${engine}
                    Log    Scanning with engine: ${engine}...
                    Predict    invoice.png
                    ${data} =    Get Result
                    Log List    ${data}
                END

        **Example: Python**

        .. code-block:: python

            lib_docai.switch_engine("base64ai")
            lib_docai.predict("invoice.png")
        """
        name: EngineName = (
            name if isinstance(name, EngineName) else EngineName(name.lower())
        )
        if name not in self._engines:
            raise RuntimeError(
                f"can't switch to {name.value!r} engine, please run"
                f" `Init Engine    {name.value}    ...` first"
            )

        self._active_engine = name

    switch_engine.__doc__ %= ", ".join([engine.value for engine in EngineName])

    @keyword
    def init_engine(
        self,
        name: Union[EngineName, str],
        secret: Optional[SecretType] = None,
        vault: Optional[Union[Dict, str]] = None,
        **kwargs,
    ) -> None:
        """Initialize the engine you want to scan documents with.

        This is required before being able to run ``Predict``. Once initialized, you
        don't need to run this again, simply use ``Switch Engine`` to jump between
        the engines. The final secret value (passed directly with `secret` or picked up
        automatically from the Vault with `vault`) will be split into authorization
        args and kwargs or just passed as it is to the wrapped library. Keep in mind
        that some engines are expecting API keys where others tokens or private keys.
        Any optional keyword argument will be passed further in the wrapped library.

        :param name: Name of the engine.
        :param secret: Authenticate with a string/file/object secret directly.
        :param vault: Specify the Vault storage `name` and secret `key` in order to
            authenticate. ('name:key' or {name: key} formats are supported)

        **How secret resolution works**

        When `vault` is passed in, the corresponding Vault is retrieved and the value
        belonging to specified field is returned as a secret. If a `secret` is used,
        then this value is returned as it is if this isn't a path pointing to the file
        holding the value to be returned. We'll be relying on environment variables in
        the absence of both the `secret` and `vault`.

        Expected secret value formats:

        - google: `<json-service/token>` (``RPA.Cloud.Google.Init Document AI``)
        - base64ai: `<e-mail>,<api-key>`
          (``RPA.DocumentAI.Base64AI.Set Authorization``)
        - nanonets: `<api-key>` (``RPA.DocumentAI.Nanonets.Set Authorization``)

        **Example: Robot Framework**

        .. code-block:: robotframework

            *** Keywords ***
            Init Base64
                Init Engine    base64ai    vault=document_ai:base64ai

        **Example: Python**

        .. code-block:: python

            from RPA.DocumentAI import DocumentAI
            from RPA.Robocorp.Vault import Vault

            lib_docai = DocumentAI()
            mail_apikey = Vault().get_secret("document_ai")["base64ai"]
            lib_docai.init_engine("base64ai", secret=mail_apikey)
        """
        if secret and vault:
            raise ValueError("choose between `secret` and `vault`")
        elif not (secret or vault):
            self.logger.warning("No `secret` or `vault` provided, relying on env vars.")
        if isinstance(vault, str):
            vault_name, vault_secret_key = vault.split(":")
            vault = {vault_name: vault_secret_key}

        init_map = {
            # Google library needs to be Vault aware due to its internal way of
            #  handling secrets directly from there. (without our parsing)
            EngineName.GOOGLE: functools.partial(self._init_google, vault=vault),
            # Rest of the engines have a normalized way of picking up secrets. They can
            #  be provided directly (string, file path, list or dict) or through Vault.
            EngineName.BASE64: self._init_base64,
            EngineName.NANONETS: self._init_nanonets,
        }
        name: EngineName = (
            name if isinstance(name, EngineName) else EngineName(name.lower())
        )
        secret_value = self._get_secret_value(secret, vault)

        # Will raise by itself if additional `kwargs` are required but not provided,
        #  given the selected engine.
        init_map[name](secret_value, **kwargs)
        self.switch_engine(name)

    @keyword
    def predict(
        self,
        location: PathType,
        model: Optional[Union[str, List[str]]] = None,
        **kwargs,
    ) -> None:
        """Scan a document with the currently active engine and store the result
        internally for a later retrieval.

        Based on the selected engine, this wraps a chain of libraries until calling a
        service API in the end, where the passed file is analyzed. Any optional keyword
        argument will be passed further in the wrapped library. (some engines require
        mandatory parameters like project ID or region)

        :param location: Path to a local file or URL address of a remote one. (not all
            engines work with URLs)
        :param model: Model name(s) to scan with. (some engines guess the model if
            not specified)

        **Example: Robot Framework**

        .. code-block:: robotframework

            *** Tasks ***
            Document AI Base64
                [Setup]    Init Base64
                Predict    https://site.com/path/to/invoice.png

        **Example: Python**

        .. code-block:: python

            lib_docai.predict("local/path/to/invoice.png", model="finance/invoice")
        """
        location_file = Path(location).expanduser().resolve()
        if location_file.exists():
            location_url = None
            self.logger.info("Using file path based location: %s", location_file)
        else:
            location_url = location
            location_file = None
            self.logger.info("Using URL address based location: %s", location_url)
            if self._active_engine in (EngineName.GOOGLE, EngineName.NANONETS):
                self.logger.warning(
                    f"Engine {self._active_engine} isn't supporting URL input at the "
                    "moment!"
                )
        if not model and self._active_engine in (
            EngineName.GOOGLE,
            EngineName.NANONETS,
        ):
            self.logger.warning(
                f"Engine {self._active_engine} requires a specific `model` passed in!"
            )

        process_map = {
            EngineName.GOOGLE: lambda: self.engine.process_document(
                file_path=location_file, processor_id=model, **kwargs
            ),
            EngineName.BASE64: lambda: (
                self.engine.scan_document_file
                if location_file
                else self.engine.scan_document_url
            )(location_file or location_url, model_types=model, **kwargs),
            EngineName.NANONETS: lambda: self.engine.predict_file(
                filepath=location_file, model_id=model
            ),
        }
        result = process_map[self._active_engine]()
        self._results[self._active_engine] = result

    @keyword
    def get_result(self, extended: bool = False) -> ResultType:
        """Retrieve the result data previously obtained with ``Predict``.

        The stored raw result is usually pre-processed with a library specific keyword
        prior the return.

        :param extended: Get all the details inside the result data. (main fields only
            by default)
        :returns: Usually a list of fields detected in the document.

        **Example: Robot Framework**

        .. code-block:: robotframework

            *** Tasks ***
            Scan With Base64
                Document AI Base64
                ${data} =    Get Result
                Log List    ${data}

        **Example: Python**

        .. code-block:: python

            result = lib_docai.get_result()
            for field in result:
                print(field)
        """
        if extended:
            return self.result

        result_map = {
            EngineName.GOOGLE: lambda: self.engine.get_document_entities(self.result),
            EngineName.BASE64: lambda: self.engine.get_fields_from_prediction_result(
                self.result
            ),
            EngineName.NANONETS: lambda: self.engine.get_fields_from_prediction_result(
                self.result
            ),
        }
        return result_map[self._active_engine]()



"""Testing of RPA.DocumentAI library."""


import json
from pathlib import Path

import pytest
from unittest import mock

from RPA.DocumentAI import DocumentAI
from RPA.DocumentAI.Base64AI import Base64AI
from RPA.DocumentAI.DocumentAI import EngineName
from RPA.DocumentAI.Nanonets import Nanonets

from . import RESOURCES_DIR, RESULTS_DIR


DOCAI_DIR = RESOURCES_DIR / "documentai"


class TestDocumentAI:
    """Tests `RPA.DocumentAI`."""

    @staticmethod
    @pytest.fixture
    def library():
        return DocumentAI()

    @staticmethod
    @pytest.fixture
    def invoice_png():
        return RESOURCES_DIR / "invoice.png"

    @pytest.mark.xfail(reason="`rpaframework-google` is unexpectedly available in venv")
    def test_init_engine_google(self, library):
        # As `rpaframework-google` isn't installed by default.
        with pytest.raises(ImportError):
            library.init_engine("google")

    def test_init_engine_base64(self, library):
        secret = "cosmin@robocorp.com,api-key"
        with pytest.raises(TypeError):
            # Two arguments (positional or keyword) are required.
            library.init_engine("base64ai", secret="api-key")
        library.init_engine("base64ai", secret=secret)
        engine = library.engine
        assert isinstance(engine, Base64AI)
        assert secret.replace(",", ":") in engine._request_headers["Authorization"]

    def test_init_engine_nanonets(self, library):
        secret = "api-key"
        library.init_engine("nanonets", secret=secret)
        engine = library.engine
        assert isinstance(engine, Nanonets)
        assert secret == library.engine.apikey

    def test_switch_engine(self, library):
        with pytest.raises(RuntimeError):
            assert library.engine  # no engine initialized

        library.init_engine("base64ai", secret="cosmin@robocorp.com,api-key")
        library.init_engine("nanonets", secret="api-key")

        assert isinstance(library.engine, Nanonets)
        library.switch_engine(EngineName.BASE64)
        assert isinstance(library.engine, Base64AI)

    @mock.patch("RPA.DocumentAI.Base64AI.requests")
    def test_predict_base64(self, mock_requests, library, invoice_png):
        library.init_engine("base64ai", secret="cosmin@robocorp.com,api-key")
        model = "invoice-model"
        library.predict(invoice_png, model=model)
        args, kwargs = mock_requests.post.call_args
        assert args == ("https://base64.ai/api/scan",)
        assert model in kwargs["json"]["modelTypes"]

    @mock.patch("RPA.DocumentAI.Nanonets.requests")
    def test_predict_nanonets(self, mock_requests, library, invoice_png):
        library.init_engine("nanonets", secret="api-key")
        model = "invoice-model"
        library.predict(invoice_png, model=model)
        args, kwargs = mock_requests.post.call_args
        assert args == (
            "https://app.nanonets.com/api/v2/OCR/Model/invoice-model/LabelFile/",
        )
        assert kwargs["files"]["file"].name == str(invoice_png)

    @pytest.mark.parametrize("extended", [False, True])
    def test_get_result(self, library, extended):
        library.init_engine("nanonets", secret="api-key")
        prediction = {
            "result": [
                {
                    "prediction": [
                        {
                            "id": "251fed51-e1fa-477e-af9d-3fe115486290",
                            "label": "bank_name",
                            "xmin": 901,
                            "ymin": 1619,
                            "xmax": 1001,
                            "ymax": 1629,
                            "score": 0.5428322,
                            "ocr_text": "Deutsche Bank",
                            "type": "field",
                            "status": "correctly_predicted",
                            "page_no": 0,
                            "label_id": "d5513122-0a46-4927-b923-cd9e6eed758d",
                        },
                        {"extra": "field"},
                    ]
                }
            ]
        }
        library._results[library._active_engine] = prediction
        result = library.get_result(extended=extended)
        if extended:
            assert result == prediction
        else:
            assert result[0]["ocr_text"] == "Deutsche Bank"


class TestBase64AI:
    """Tests `RPA.DocumentAI.Base64AI`."""

    @staticmethod
    @pytest.fixture
    def library():
        return Base64AI()

    @staticmethod
    @pytest.fixture
    def input_files():
        paths = {
            "driving-license": DOCAI_DIR / "signature-license.jpg",
            "payment-check": DOCAI_DIR / "signature-check.png",
        }
        urls = {
            "driving-license": (
                "https://raw.githubusercontent.com/robocorp/rpaframework/5d8f3db0e3fe5"
                "5396afccd987afaa1f389df2527/packages/main/tests/resources/documentai/"
                "signature-license.jpg"
            ),
            "payment-check": (
                "https://raw.githubusercontent.com/robocorp/rpaframework/5d8f3db0e3fe5"
                "5396afccd987afaa1f389df2527/packages/main/tests/resources/documentai/"
                "signature-check.png"
            ),
        }
        return {
            "paths": paths,
            "urls": urls,
        }

    @staticmethod
    @pytest.fixture
    def signature_response():
        with open(DOCAI_DIR / "signature-recognize-response.json") as stream:
            return json.load(stream)

    @pytest.mark.parametrize(
        "source,field_ending",
        [
            # source type, reference source ending, query source ending
            ("paths", ("2Q==", "gg==")),
            ("urls", ("signature-license.jpg", "signature-check.png")),
        ],
    )
    @mock.patch("RPA.DocumentAI.Base64AI.requests")
    def test_get_matching_signatures(
        self, mock_requests, library, input_files, source, field_ending
    ):
        files = input_files[source]
        _ = library.get_matching_signatures(
            files["driving-license"], files["payment-check"]
        )
        args, kwargs = mock_requests.post.call_args
        assert args == ("https://base64.ai/api/signature",)
        payload = kwargs["json"]
        if source == "paths":
            ref_key, qry_key = "document", "queryDocument"
        else:
            ref_key, qry_key = "url", "queryUrl"
        assert payload[ref_key].endswith(field_ending[0])
        assert payload[qry_key].endswith(field_ending[1])

    @pytest.mark.parametrize(
        "threshold,ref_idx,qry_idxes,similarities",
        [
            (0.8, 0, [0], [0.89]),  # only one accepted match for the first reference
            (0.5, 0, [0], [0.89]),  # still the same even when lowering thresholds
            (0.5, 1, [0, 1], [0.66, 0.61]),  # but different for the 2nd reference
        ],
    )
    def test_filter_matching_signatures(
        self, library, signature_response, threshold, ref_idx, qry_idxes, similarities
    ):
        matches = library.filter_matching_signatures(
            signature_response,
            confidence_threshold=threshold,
            similarity_threshold=threshold,
        )
        ref_key = [key for key in matches.keys() if key[0] == ref_idx][0]
        qry_imgs = matches[ref_key]
        assert len(qry_imgs) == len(qry_idxes) == len(similarities)

        for qry_idx, similarity in zip(qry_idxes, similarities):
            qry_img = [img for img in qry_imgs if img["index"] == qry_idx][0]
            assert qry_img["similarity"] == similarity

    @pytest.mark.parametrize(
        "index,reference,path,name,subcontent",
        [
            (0, False, None, "query-0.png", b"@@@@@@@@@@@@@@@@@@@@@@"),
            (0, True, RESULTS_DIR / "ref-sig.jpg", "ref-sig.jpg", b"(((((((((((((((("),
        ],
    )
    @mock.patch("RPA.DocumentAI.Base64AI.get_output_dir")
    def test_get_signature_image(
        self,
        mock_get_output_dir,
        library,
        signature_response,
        index,
        reference,
        path,
        name,
        subcontent,
    ):
        mock_get_output_dir.return_value = RESULTS_DIR
        image_path_str = library.get_signature_image(
            signature_response, index=index, reference=reference, path=path
        )
        image_path = Path(image_path_str)
        assert image_path.name == name
        assert subcontent in image_path.read_bytes()



#!/usr/bin/env python

#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

# NOTE(cmin764; 7 Sep 2023): Vendorizing `robotframework`'s entry-point in order to
#  execute the SSL injection (and ony other runtime patch) as soon as possible, way
#  before importing any other package, thus risking failing the injection.
# pylint: disable=unused-import
import RPA.core  # noqa: F401

import sys

if __name__ == "__main__" and "robot" not in sys.modules:
    # pylint: disable=unused-import
    import pythonpathsetter  # noqa: F401

from robot import run_cli

run_cli()



import base64
import itertools
import logging
import mimetypes
import urllib.parse as urlparse
from typing import Any, Dict, List, Optional, Tuple, Union

import requests
import validators

from RPA.JSON import JSONType
from RPA.Robocorp.utils import PathType, get_output_dir
from RPA.RobotLogListener import RobotLogListener


class Base64AI:
    """Library to support `Base64.ai <https://base64.ai/>`_ service for intelligent
    document processing (IDP).

    Library requires at the minimum `rpaframework` version **19.0.0**.

    Service supports identifying fields in the documents, which can be given to the
    service in multiple different file formats and via URL.

    **Robot Framework example usage**

    .. code-block:: robotframework

        *** Settings ***
        Library   RPA.DocumentAI.Base64AI
        Library   RPA.Robocorp.Vault

        *** Tasks ***
        Identify document
            ${secrets}=   Get Secret  base64ai-auth
            Set Authorization  ${secrets}[email-address]   ${secrets}[apikey]
            ${results}=  Scan Document File
            ...   ${CURDIR}${/}invoice.pdf
            ...   model_types=finance/check/usa,finance/invoice/usa
            # Scan response contains list of detected models in the document
            FOR  ${result}  IN  @{results}
                Log To Console  Model: ${result}[model]
                Log To Console  Field keys: ${{','.join($result['fields'].keys())}}
                Log To Console  Fields: ${result}[fields]
                Log To Console  Text (OCR): ${result}[ocr]
            END


    **Python example usage**

    .. code-block:: python

        from RPA.DocumentAI.Base64AI import Base64AI
        from RPA.Robocorp.Vault import Vault

        secrets = Vault().get_secret("base64ai-auth")
        baselib = Base64AI()
        baselib.set_authorization(secrets["email-address"], secrets["apikey"])
        result = baselib.scan_document_file(
            "invoice.pdf",
            model_types="finance/invoice,finance/check/usa",
        )
        for r in result:
            print(f"Model: {r['model']}")
            for key, props in r["fields"].items():
                print(f"FIELD {key}: {props['value']}")
            print(f"Text (OCR): {r['ocr']}")

    Portal example: https://github.com/robocorp/example-idp-base64
    """

    ROBOT_LIBRARY_SCOPE = "GLOBAL"
    ROBOT_LIBRARY_DOC_FORMAT = "REST"

    BASE_URL = "https://base64.ai"

    def __init__(self):
        self.logger = logging.getLogger(__name__)
        self._request_headers = {"Content-Type": "application/json"}
        listener = RobotLogListener()
        listener.register_protected_keywords(
            ["RPA.DocumentAI.Base64AI.set_authorization"]
        )

    def set_authorization(self, api_email: str, api_key: str) -> None:
        """Set Base64 AI request headers with email and key related to API.

        :param api_email: email address related to the API
        :param api_key: key related to the API

        Robot Framework example:

        .. code-block:: robotframework

            ${secrets}=   Get Secret  base64ai-auth
            Set Authorization    ${secrets}[email-address]    ${secrets}[apikey]

        Python example:

        .. code-block:: python

            secrets = Vault().get_secret("base64ai-auth")
            baselib = Base64AI()
            baselib.set_authorization(secrets["email-address"], secrets["apikey"])
        """
        self._request_headers["Authorization"] = f"ApiKey {api_email}:{api_key}"

    @classmethod
    def _to_endpoint(cls, part: str, mock: bool = False) -> str:
        return urlparse.urljoin(
            cls.BASE_URL, f"{'mock' if mock else 'api'}/{part.strip('/')}"
        )

    def _scan_document(
        self,
        payload: Dict,
        model_types: Optional[Union[str, List[str]]] = None,
        mock: bool = False,
    ) -> JSONType:
        scan_endpoint = self._to_endpoint("scan", mock=mock)
        self.logger.info(f"Endpoint {scan_endpoint!r} is set for scanning.")
        if model_types:
            req_model_types = (
                model_types if isinstance(model_types, list) else model_types.split(",")
            )
            payload["modelTypes"] = req_model_types

        response = requests.post(  # pylint: disable=missing-timeout
            scan_endpoint,
            headers=self._request_headers,
            json=payload,
        )
        response.raise_for_status()
        return response.json()

    @staticmethod
    def _get_file_base64_and_mimetype(file_path: str) -> Tuple[str, str]:
        with open(file_path, "rb") as image_file:
            encoded_content = base64.b64encode(image_file.read())
        return encoded_content.decode("utf-8"), mimetypes.guess_type(file_path)[0]

    @classmethod
    def _url_or_file(cls, resource: str, data_only: bool = False) -> Tuple[str, bool]:
        # Returns the URL string (or file API-ready content) and resource type.
        #  (`True` if URL)
        if not data_only and validators.url(resource):
            return resource, True

        # We're dealing with a file then.
        base64string, mime = cls._get_file_base64_and_mimetype(resource)
        data = f"data:{mime};base64,{base64string}"
        return data, False

    def scan_document_file(
        self,
        file_path: str,
        model_types: Optional[Union[str, List[str]]] = None,
        mock: bool = False,
    ) -> JSONType:
        """Scan a document file. Can be given a ``model_types`` to
        specifically target certain models.

        :param file_path: filepath to the file
        :param model_types: single model type or list of model types
        :param mock: set to True to use /mock/scan endpoint instead of /scan
        :return: result of the document scan

        Robot Framework example:

        .. code-block:: robotframework

            ${results}=    Scan Document File
            ...    ${CURDIR}${/}files${/}IMG_8277.jpeg
            ...    model_types=finance/check/usa,finance/invoice
            FOR    ${result}    IN    @{results}
                Log To Console    Model: ${result}[model]
                Log To Console    Fields: ${result}[fields]
                Log To Console    Text (OCR): ${result}[ocr]
            END

        Python example:

        .. code-block:: python

            result = baselib.scan_document_file(
                "./files/Invoice-1120.pdf",
                model_types="finance/invoice,finance/check/usa",
            )
            for r in result:
                print(f"Model: {r['model']}")
                for key, val in r["fields"].items():
                    print(f"{key}: {val['value']}")
                print(f"Text (OCR): {r['ocr']}")
        """
        data, _ = self._url_or_file(file_path, data_only=True)
        payload = {"image": data}
        return self._scan_document(payload, model_types=model_types, mock=mock)

    def scan_document_url(
        self,
        url: str,
        model_types: Optional[Union[str, List[str]]] = None,
        mock: bool = False,
    ) -> JSONType:
        """Scan a document URL. Can be given a ``model_types`` to
        specifically target certain models.

        :param url: valid url to a file
        :param model_types: single model type or list of model types
        :param mock: set to True to use /mock/scan endpoint instead of /scan
        :return: result of the document scan

        Robot Framework example:

        .. code-block:: robotframework

            ${results}=    Scan Document URL
            ...    https://base64.ai/static/content/features/data-extraction/models//2.png
            FOR    ${result}    IN    @{results}
                Log To Console    Model: ${result}[model]
                Log To Console    Fields: ${result}[fields]
                Log To Console    Text (OCR): ${result}[ocr]
            END

        Python example:

        .. code-block:: python

            result = baselib.scan_document_url(
                "https://base64.ai/static/content/features/data-extraction/models//2.png"
            )
            for r in result:
                print(f"Model: {r['model']}")
                for key, props in r["fields"].items():
                    print(f"FIELD {key}: {props['value']}")
                print(f"Text (OCR): {r['ocr']}")
        """  # noqa: E501
        payload = {"url": url}
        return self._scan_document(payload, model_types=model_types, mock=mock)

    def get_fields_from_prediction_result(self, prediction: JSONType) -> List:
        """Helper keyword to get found fields from a prediction result.
        For example see ``Scan Document File`` or ``Scan Document URL`` keyword.

        :param prediction: prediction result dictionary
        :return: list of found fields
        """
        return list(
            itertools.chain(*(list(item["fields"].values()) for item in prediction))
        )

    def get_user_data(self) -> Dict:
        """Get user data including details on credits used and credits remaining
        for the Base64 service.

        Returned user data contains following keys:

            - givenName
            - familyName
            - email
            - hasWorkEmail
            - companyName
            - numberOfCredits
            - numberOfPages
            - numberOfUploads
            - numberOfCreditsSpentOnDocuments (visible if used)
            - numberOfCreditsSpentOnFaceDetection (visible if used)
            - numberOfCreditsSpentOnFaceRecognition (visible if used)
            - hasActiveAwsContract
            - subscriptionType
            - subscriptionPeriod
            - tags
            - ccEmails
            - status
            - remainingCredits (calculated by the keyword)

        :return: object containing details on the API user

        Robot Framework example:

        .. code-block:: robotframework

            ${userdata}=   Get User Data
            Log To Console  I have still ${userdata}[remainingCredits] credits left

        Python example:

        .. code-block:: python

            userdata = baselib.get_user_data()
            print(f"I have still {userdata['remainingCredits']} credits left")
        """
        response = requests.get(  # pylint: disable=missing-timeout
            self._to_endpoint("auth/user"),
            headers=self._request_headers,
        )
        response.raise_for_status()
        json_response = response.json()
        spent_on_documents = json_response.get("numberOfCreditsSpentOnDocuments", 0)
        spent_on_face_detection = json_response.get(
            "numberOfCreditsSpentOnFaceDetection", 0
        )
        spent_on_face_recognition = json_response.get(
            "numberOfCreditsSpentOnFaceRecognition", 0
        )
        remainingCredits = (
            json_response["numberOfCredits"]
            - spent_on_documents
            - spent_on_face_detection
            - spent_on_face_recognition
        )
        json_response["remainingCredits"] = remainingCredits
        return json_response

    def get_matching_signatures(
        self, reference_image: PathType, query_image: PathType
    ) -> JSONType:
        """Returns a list of matching signatures found from the reference into the
        queried image.

        The input images can be paths to the files or URLs.

        The output JSON-like dictionary contains all the details from the API, like the
        detected signatures in both the reference and query image and for every such
        signature, its bounding-box geometry, confidence and similarity score.
        Use the ``Filter Matching Signatures`` over this value to get a simpler
        structure.

        :param reference_image: The reference image (jpg/png) to check query signatures
            against. (e.g. driving license, ID card)
        :param query_image: The query image containing signatures similar to the ones
            from the reference image. (e.g. signed contract, bank check)
        :returns: A JSON-like dictionary revealing recognized signatures and how much
            they resemble with each other.

        **Example: Robot Framework**

        .. code-block:: robotframework

            *** Tasks ***
            Match Signatures
                ${ref_image} =  Set Variable    driving-license.jpg
                ${query_image} =  Set Variable    signed-check.png
                ${sigs} =   Get Matching Signatures     ${ref_image}    ${query_image}

        **Example: Python**

        .. code-block:: python

            from RPA.DocumentAI.Base64AI import Base64AI

            lib = Base64AI()
            sigs = lib.get_matching_signatures(
                "driving-license.jpg", "signed-check.png"
            )

        Portal example: https://github.com/robocorp/example-signature-match-assistant
        """
        # NOTE(cmin764): There's no mock support for this API.
        recognize_endpoint = self._to_endpoint("signature")

        reference, query = map(
            lambda img: self._url_or_file(str(img)), (reference_image, query_image)
        )
        payload = {
            f"{'url' if reference[1] else 'document'}": reference[0],
            f"{'queryUrl' if query[1] else 'queryDocument'}": query[0],
        }

        response = requests.post(  # pylint: disable=missing-timeout
            recognize_endpoint,
            headers=self._request_headers,
            json=payload,
        )
        response.raise_for_status()
        match_response = response.json()
        return match_response

    def filter_matching_signatures(
        self,
        match_response: JSONType,
        confidence_threshold: float = 0.8,
        similarity_threshold: float = 0.8,
    ) -> Dict[Tuple[int, Tuple[int, ...]], List[Dict[str, Any]]]:
        """Gets through all the recognized signatures in the queried image and returns
        only the ones passing the confidence & similarity thresholds.

        Additionally, this keyword simplifies the original input `match_response`
        structure and returns a dictionary with all the detected and accepted reference
        signatures as keys, and lists of similar enough query signatures as values.

        - Each reference signature (key) is a tuple of `(index, coordinates)`.
        - Each query signature (sub-value) is a dictionary of `{index, coords,
          similarity}`.
        - The coordinates describe the bounding-box enclosing the detected signature
          portion from the original image, as follows: `(left, top, right, bottom)`
          corners.

        Use the original `match_response` object and the indexes from here if you need
        to retrieve extra details not found here (e.g. confidence score). Use the
        ``Get Signature Image`` to save and preview the image crop belonging to the
        signature of choice.

        :param match_response: The raw JSON-like response retrieved with the
            ``Get Matching Signatures`` keyword.
        :param confidence_threshold: The minimum accepted confidence score (0.0-1.0)
            for a candidate to be considered a signature. (to avoid false-positives)
        :param similarity_threshold: The minimum accepted similarity score (0.0-1.0)
            for a query signature to be considered an alike signature. (to discard
            different or fraudulent signatures)
        :returns: A dictionary of accepted reference signatures and their similar ones
            found in the queried image.

        **Example: Robot Framework**

        .. code-block:: robotframework

            *** Tasks ***
            Match Signatures
                &{matches} =   Filter Matching Signatures      ${sigs}
                Log Dictionary    ${matches}

        **Example: Python**

        .. code-block:: python

            matches = lib.filter_matching_signatures(sigs)
            print(matches)
        """
        accepted_references = [
            ref["confidence"] >= confidence_threshold
            for ref in match_response["reference"]
        ]
        candidates = {}
        # As (left, top, right, bottom) tuple.
        to_coords = (
            lambda item: (  # noqa: E731, pylint: disable=unnecessary-lambda-assignment
                item["left"],
                item["top"],
                item["left"] + item["width"],  # right corner
                item["top"] + item["height"],  # bottom corner
            )
        )

        for qry_idx, candidate in enumerate(match_response["query"]):
            if candidate["confidence"] < confidence_threshold:
                continue  # not sure enough it is a signature

            qry_coords = to_coords(candidate)
            for ref_idx, similarity in enumerate(candidate["similarities"]):
                if (
                    not accepted_references[ref_idx]
                    or similarity < similarity_threshold
                ):
                    # Skip any resemblance to doubtful reference signatures.
                    # Skip not similar enough signatures.
                    continue

                # Filtered properties for the accepted match.
                qry_body = {
                    "index": qry_idx,
                    "coords": qry_coords,
                    "similarity": similarity,
                }
                # Every accepted reference signature (index, coordinates) contains a
                #  list of similar recognized signatures from the queried image, as
                #  {index, coordinates, similarity}.
                ref_coords = to_coords(match_response["reference"][ref_idx])
                # Example of such candidate dictionary entry:
                #  (0, (92, 509, 234, 616)): [{'index': 0, 'coords': (1812, 813, 2114, 1035), 'similarity': 0.89}]  # noqa: E501
                candidates.setdefault((ref_idx, ref_coords), []).append(qry_body)

        for matches in candidates.values():
            matches.sort(key=lambda body: body["similarity"], reverse=True)
        return candidates

    def get_signature_image(
        self,
        match_response: JSONType,
        *,
        index: int,
        reference: bool = False,
        path: Optional[PathType] = None,
    ) -> str:
        """Retrieves and saves locally the image cut belonging to the provided `index`.

        The image data itself is provided with the original `match_response` object as
        base64 encoded content. This utility keyword retrieves, decodes and saves it
        on the local disk customized with the `path` parameter. By default, the
        searched `index` is considered a query image, switch to the reference type by
        enabling it with the `reference` parameter.

        :param match_response: The raw JSON-like response retrieved with the
            ``Get Matching Signatures`` keyword.
        :param index: The image ID (numeric) found along the coordinates in the output
            of the ``Filter Matching Signatures`` keyword. (the list order is stable)
        :param reference: Set this to `True` if you're looking for a reference (not
            query) image instead. (off by default)
        :param path: Set an explicit output path (including file name) for the locally
            saved image. (uses the output directory as default)
        :returns: The image path of the locally saved file.

        **Example: Robot Framework**

        .. code-block:: robotframework

            *** Tasks ***
            Match Signatures
                @{ref_sigs} =   Get Dictionary Keys    ${matches}
                @{qry_sigs} =    Get From Dictionary    ${matches}    ${ref_sigs}[${0}]
                &{qry_sig} =    Set Variable    ${qry_sigs}[${0}]
                ${path} =   Get Signature Image     ${sigs}     index=${qry_sig}[index]
                Log To Console    Preview query signature image crop: ${path}

        **Example: Python**

        .. code-block:: python

            qry_sig = list(matches.values())[0][0]
            path = lib.get_signature_image(sigs, index=qry_sig["index"])
            print("Preview query signature image crop: ", path)
        """
        images_type = "reference" if reference else "query"
        images = match_response[images_type]
        encoded_content = images[index]["image"]
        image_format, image_content = encoded_content.split(";base64,")
        decoded_bytes = base64.b64decode(image_content.encode())
        image_ext = image_format.split("/")[-1]
        path = path or (get_output_dir() / f"{images_type}-{index}.{image_ext}")
        with open(path, "wb") as stream:
            stream.write(decoded_bytes)
        return str(path)



import pytest
from RPA.Calendar import Calendar


@pytest.fixture
def library():
    return Calendar()


def test_returning_holidays(library):
    holidays = library.return_holidays(2023, "FI")
    for date, holiday_name in holidays.items():
        print(f"{date} is {holiday_name}")
    assert len(holidays) == 15


@pytest.mark.freeze_time("2023-03-09")
def test_set_locale_spanish(library):
    library.set_locale("es")
    now = library.time_now(
        timezone="Europe/Helsinki", return_format="dddd DD MMMM YYYY"
    )
    assert now == "jueves 09 marzo 2023"


@pytest.mark.freeze_time("2023-03-09")
def test_set_locale_english(library):
    library.set_locale("en")
    now = library.time_now(
        timezone="Europe/Helsinki", return_format="dddd DD MMMM YYYY"
    )
    assert now == "Thursday 09 March 2023"



import platform
import time


class Buffer:
    """A context manager that waits for a guaranteed minimum interval
    between the previous exit and next enter.
    """

    def __init__(self, logger, interval=0.1):
        self.logger = logger
        self.default = float(interval)
        self.interval = self.default
        self.previous = 0.0

    def __call__(self, interval=None):
        if interval is not None:
            self.interval = float(interval)

        return self

    def __enter__(self):
        duration = time.time() - self.previous
        if duration < self.interval:
            sleep_time = self.interval - duration
            self.logger.debug("Buffering input for %.2f seconds", sleep_time)
            time.sleep(sleep_time)

        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.previous = time.time()
        self.interval = self.default


def is_windows():
    return platform.system() == "Windows"


def is_macos():
    return platform.system() == "Darwin"


def is_linux():
    return platform.system() == "Linux"



"""Base utilities for COM applications like Word, Excel, Outlook."""

import atexit
import gc
import logging
import platform
import struct
from contextlib import contextmanager
from pathlib import Path

if platform.system() == "Windows":
    import win32api
    import win32com.client
    from pywintypes import com_error as COMError  # pylint: disable=no-name-in-module
    from win32com.client import constants  # pylint: disable=unused-import
    import win32com
else:
    # Define dummy objects for non-Windows platforms
    win32api = None
    win32com = None
    COMError = Exception
    constants = None
    logging.getLogger(__name__).warning(
        "Any `RPA.*.Application` library works only on Windows platform!"
    )
    # As these are imported anyway from here and should be defined on non-Windows OSes.
    COMError = Exception
    constants = None


def rgb_to_excel_color(red, green, blue):
    return red + (green * 256) + (blue * 256**2)


def _to_unsigned(val):
    return struct.unpack("L", struct.pack("l", val))[0]


def to_path(path: str) -> Path:
    return Path(path).expanduser().resolve()


def to_str_path(path: str) -> str:
    return str(to_path(path))


# TODO(cmin764; 16 Aug 2023): Create a `handles_com_error` decorator if we get a little
#  too repetitive with context managers guarding. (to decorate the keywords)
@contextmanager
def catch_com_error():
    """Try to convert COM errors to human-readable format."""
    try:
        yield
    except COMError as err:  # pylint: disable=no-member
        if err.excepinfo:
            try:
                msg = win32api.FormatMessage(_to_unsigned(err.excepinfo[5]))
            except Exception:  # pylint: disable=broad-except
                msg = err.excepinfo[2]
        else:
            try:
                msg = win32api.FormatMessage(_to_unsigned(err.hresult))
            except Exception:  # pylint: disable=broad-except
                msg = err.strerror
        raise RuntimeError(msg) from err


class MetaApplication(type):
    """Metaclass enhancing every final class inheriting from a base using this."""

    def __new__(cls, *args, **kwargs) -> type:
        final_class = super().__new__(cls, *args, **kwargs)
        bases = final_class.__mro__[1:]
        if len(bases) < 2:  # not enhancing bases
            return final_class

        super_class = bases[0]
        final_class.__doc__ = (final_class.__doc__ or "") + (super_class.__doc__ or "")
        if final_class.APP_DISPATCH is None:
            raise ValueError(
                "An `APP_DISPATCH` has to be defined in this"
                f" {final_class.__name__!r} class"
            )
        return final_class


class BaseApplication(metaclass=MetaApplication):
    # Base library `Application` class for managing COM apps.
    # The docstring below is automatically appended at the end of every inheritor.
    """
    **Caveats**

    This library works on a Windows operating system with UI enabled only, and you must
    ensure that you open the app first with ``Open Application`` before running any
    other relevant keyword which requires to operate on an open app. The application is
    automatically closed at the end of the task execution, so this can be changed by
    importing the library with the `autoexit=${False}` setting.

    .. code-block:: robotframework

        *** Settings ***
        Library     RPA.Excel|Outlook|Word.Application    autoexit=${False}

    If you're running the Process by Control Room through a custom self-hosted Worker
    service, then please make sure that you enable an RDP session by ticking "Use
    Desktop Connection" under the Step configuration.

    If you still encounter issues with opening a document, please ensure that file can
    be opened first manually and dismiss any alert potentially blocking the process.

    Check the documentation below for more info:

    - https://robocorp.com/docs/control-room/unattended/worker-setups/windows-desktop
    - https://robocorp.com/docs/faq/windows-server-2016
    """

    ROBOT_LIBRARY_SCOPE = "GLOBAL"
    ROBOT_LIBRARY_DOC_FORMAT = "REST"

    APP_DISPATCH = None  # this has to be defined in every inheriting class

    def __init__(self, autoexit: bool = True):
        """Initialize the library instance by wrapping the COM Windows app.

        :param autoexit: Automatically close the app when the process finishes.
        """
        self.logger = logging.getLogger(__name__)
        self._app_name = self.APP_DISPATCH.split(".")[0]
        self._app = None

        if platform.system() != "Windows":
            self.logger.warning(
                "This %s application library requires Windows dependencies to work.",
                self._app_name,
            )
        if autoexit:
            atexit.register(self.quit_application)

    @property
    def app(self):
        if self._app is None:
            raise ValueError(f"{self._app_name} application is not open")

        return self._app

    def open_application(
        self, visible: bool = False, display_alerts: bool = False
    ) -> None:
        """Open the application.

        :param visible: Show the window on opening. (`False` by default)
        :param display_alerts: Display alert popups. (`False` by default)
        """
        with catch_com_error():
            self._app = win32com.client.gencache.EnsureDispatch(self.APP_DISPATCH)
            self.logger.debug("Opened application: %s", self.app)

            if hasattr(self.app, "Visible"):
                state = "visible" if visible else "invisible"
                self.logger.debug("Making the application %s.", state)
                self.app.Visible = visible
            elif visible:
                self.logger.warning("Visibility cannot be controlled on this app.")

            # Show for e.g. a file overwrite warning or not.
            if hasattr(self.app, "DisplayAlerts"):
                state = "Displaying" if display_alerts else "Hiding"
                self.logger.debug("%s the application alerts.", state)
                self.app.DisplayAlerts = display_alerts
            elif display_alerts:
                self.logger.warning(
                    "Alerts displaying cannot be controlled on this app."
                )

    @property
    def _active_document(self):
        # Retrieves the currently active document. (raises if there's none active)
        with catch_com_error():
            return getattr(self.app, "ActiveDocument", None)

    def _deactivate_document(self):
        # Cleans-up a just closed previously active document.
        pass

    def close_document(self, save_changes: bool = False) -> None:
        """Close the active document and app (if open).

        :param save_changes: Enable changes saving on quit. (`False` by default)
        """
        if not self._app:  # no app open at all
            return

        try:
            with catch_com_error():
                if self._active_document:
                    state = "saving" if save_changes else "not saving"
                    self.logger.debug(
                        "Closing the open document and %s changes.", state
                    )
                    self._active_document.Close(save_changes)
                    self._deactivate_document()
        except RuntimeError as exc:
            if "no document is open" in str(exc):
                self.logger.warning(
                    "Failed attempt on closing a document when there's none open!"
                )
            else:
                raise

    def quit_application(self, save_changes: bool = False) -> None:
        """Quit the application.

        :param save_changes: Enable to save changes on quit. (`False` by default)
        """
        if not self._app:
            return

        self.close_document(save_changes)
        with catch_com_error():
            gc.collect()
            self.app.Quit()
        self._app = None
        gc.collect()

    def set_object_property(self, object_instance, property_name: str, value: str):
        """Set the property of any object.

        This is a utility keyword for Robot Framework syntax to set object
        property values.

        .. code-block:: robotframework

            ${new_value}=    Replace String    ${value}    10.132.    5511.11.
            Set Object Property    ${result}    Value    ${new_value}

        :param object_instance: object instance to set the property
        :param property_name: property name to set
        :param value: value to set
        """
        if hasattr(object_instance, property_name):
            setattr(object_instance, property_name, value)
        else:
            raise AttributeError(
                f"Object {type(object_instance)} does not "
                f"have property '{property_name}'"
            )



# pylint: disable=too-many-lines
import logging
import pathlib
import re
from collections import defaultdict
from contextlib import contextmanager
from io import BytesIO
from typing import Any, List, Optional, Union

import openpyxl
import xlrd
import xlwt

from PIL import Image, ImageColor
from xlutils.copy import copy as xlutils_copy
from openpyxl.worksheet.cell_range import CellRange
from openpyxl.utils import get_column_letter, column_index_from_string
from openpyxl.utils import cell as utils_cell
from openpyxl.utils.exceptions import InvalidFileException
from openpyxl.styles import Font, PatternFill, Alignment
from openpyxl.styles.colors import Color as xlsColor
from openpyxl.formula.translate import Translator
from RPA.Tables import Table, Tables, return_table_as_raw_list


PathType = Union[str, pathlib.Path]


def get_column_index(column: str) -> int:
    """Get column index from name, e.g. A -> 1, D -> 4, AC -> 29.
    Reverse of `get_column_letter()`
    """
    column = str(column).lower()

    col = 0
    for digit, char in enumerate(column[::-1]):
        value = ord(char) - 96
        col += (26**digit) * value

    return col


def ensure_unique(values: Any) -> List[Any]:
    """Ensures that each string value in the list is unique.
    Adds a suffix to each value that has duplicates,
    e.g. [Banana, Apple, Lemon, Apple] -> [Banana, Apple, Lemon, Apple_2]
    """

    def to_unique(values: Any) -> List[Any]:
        output = []
        seen = defaultdict(int)
        for value in values:
            if seen[value] and isinstance(value, str):
                output.append("%s_%d" % (value, seen[value] + 1))
            else:
                output.append(value)
            seen[value] += 1
        return output

    # Repeat process until each column is unique
    output = to_unique(values)
    while True:
        verify = to_unique(output)
        if output == verify:
            break
        output = verify

    return output


class Files:
    """The `Excel.Files` library can be used to read and write Excel
    files without the need to start the actual Excel application.

    It supports both legacy ``.xls`` files and modern ``.xlsx`` files.

    **Note:** To run macros or load password protected worksheets,
    please use the Excel application library.

    **Examples**

    **Robot Framework**

    A common use-case is to load an existing Excel file as a table,
    which can be iterated over later in a Robot Framework keyword or task:

    .. code-block:: robotframework

        *** Settings ***
        Library    RPA.Tables
        Library    RPA.Excel.Files

        *** Keywords ***
        Read orders as table
            Open workbook    ${ORDERS_FILE}
            ${worksheet}=    Read worksheet   header=${TRUE}
            ${orders}=       Create table     ${worksheet}
            [Return]         ${orders}
            [Teardown]       Close workbook

    Processing all worksheets in the Excel file and checking row count:

    .. code-block:: robotframework

        *** Settings ***
        Library    RPA.Excel.Files

        *** Variables ***
        ${EXCEL_FILE}   /path/to/excel.xlsx

        *** Tasks ***
        Rows in the sheet
            [Setup]      Open Workbook    ${EXCEL_FILE}
            @{sheets}=   List Worksheets
            FOR  ${sheet}  IN   @{sheets}
                ${count}=  Get row count in the sheet   ${sheet}
                Log   Worksheet '${sheet}' has ${count} rows
            END

        *** Keywords ***
        Get row count in the sheet
            [Arguments]      ${SHEET_NAME}
            ${sheet}=        Read Worksheet   ${SHEET_NAME}
            ${rows}=         Get Length  ${sheet}
            [Return]         ${rows}

    Creating a new Excel file with a dictionary:

    .. code-block:: robotframework

        *** Tasks ***
        Creating new Excel
            Create Workbook  my_new_excel.xlsx
            FOR    ${index}    IN RANGE    20
                &{row}=       Create Dictionary
                ...           Row No   ${index}
                ...           Amount   ${index * 25}
                Append Rows to Worksheet  ${row}  header=${TRUE}
            END
            Save Workbook

    Creating a new Excel file with a list:

    .. code-block:: robotframework

        *** Variables ***
        @{heading}   Row No   Amount
        @{rows}      ${heading}

        *** Tasks ***
        Creating new Excel
            Create Workbook  my_new_excel.xlsx
            FOR    ${index}    IN RANGE   1  20
                @{row}=         Create List   ${index}   ${index * 25}
                Append To List  ${rows}  ${row}
            END
            Append Rows to Worksheet  ${rows}
            Save Workbook

    **Python**

    The library can also be imported directly into Python.

    .. code-block:: python

        from RPA.Excel.Files import Files

        def read_excel_worksheet(path, worksheet):
            lib = Files()
            lib.open_workbook(path)
            try:
                return lib.read_worksheet(worksheet)
            finally:
                lib.close_workbook()
    """

    ROBOT_LIBRARY_SCOPE = "GLOBAL"
    ROBOT_LIBRARY_DOC_FORMAT = "REST"

    def __init__(self):
        self.logger = logging.getLogger(__name__)
        self.workbook = None

    def _require_open_xlsx_workbook(self, keyword_name: str):
        assert self.workbook, "No active workbook"
        if isinstance(self.workbook, XlsWorkbook):
            raise NotImplementedError(
                f"Keyword '{keyword_name}' does not support XLS files"
            )

    def _load_workbook(
        self, path: str, data_only: bool, read_only: bool
    ) -> Union["XlsWorkbook", "XlsxWorkbook"]:
        # pylint: disable=broad-except
        path = pathlib.Path(path).resolve(strict=True)

        try:
            book = XlsxWorkbook(path)
            book.open(data_only=data_only, read_only=read_only)
            return book
        except InvalidFileException as exc:
            self.logger.debug(exc)  # Unsupported extension, silently try xlrd
        except Exception as exc:
            self.logger.info(
                "Failed to open as Office Open XML (.xlsx) format: %s", exc
            )

        try:
            book = XlsWorkbook(path)
            book.open()
            return book
        except Exception as exc:
            self.logger.info("Failed to open as Excel Binary Format (.xls): %s", exc)

        raise ValueError(
            f"Failed to open Excel file ({path}), "
            "verify that the path and extension are correct"
        )

    def create_workbook(
        self,
        path: Optional[str] = None,
        fmt: str = "xlsx",
        sheet_name: Optional[str] = None,
    ) -> Union["XlsWorkbook", "XlsxWorkbook"]:
        """Create and open a new Excel workbook.

        Automatically also creates a new worksheet with the name `sheet_name`.
        (defaults to "Sheet")

        **Note:** Must be paired with the ``Save Workbook`` keyword
        or the newly created workbook will be deleted upon robot completion.

        **Note:** The file name/path must be set in either the ``Create Workbook``
        keyword or the ``Save Workbook`` keyword and must include the file extension.

        :param path: Save path for workbook; defaults to robot root if not provided.
        :param fmt: Format of workbook, i.e. xlsx or xls; Defaults to xlsx if not
            provided.
        :param sheet_name: Custom name for the initial sheet.
        :return: Workbook object.

        Examples:

        .. code-block:: robotframework

            # Create modern format workbook.
            Create Workbook
            Save Workbook    orders.xlsx

            # Create modern format workbook with custom sheet name.
            Create Workbook  sheet_name=MyCustomSheetName
            Save Workbook    orders.xlsx

            # Create modern format workbook with a path set.
            Create Workbook    path=${OUTPUT_DIR}${/}orders.xlsx
            Save Workbook

            # Create legacy format workbook.
            Create Workbook    fmt=xls
            Save Workbook    orders.xls

            # Create legacy format workbook with a path set.
            # Note that the file name must be set in the `Create Workbook` keyword
            #  if the path argument is used.
            Create Workbook    path=${OUTPUT_DIR}${/}orders.xls    fmt=xls
            Save Workbook

        .. code-block:: python

            # Create modern format workbook with defaults.
            lib = Files()
            lib.create_workbook()
            lib.save_workbook("orders.xlsx")

            # Create modern format workbook with a path set.
            lib = Files()
            lib.create_workbook(path="./output/orders.xlsx", fmt="xlsx")
            lib.save_workbook()

            # Create legacy format workbook.
            lib = Files()
            lib.create_workbook(fmt="xls")
            lib.save_workbook("orders.xls")

            # Create legacy format workbook with a path set.
            # Note that the file name must be set in the `Create Workbook` keyword
            #  if the path is used.
            lib = Files()
            lib.create_workbook(path="./output/orders.xls", fmt="xls")
            lib.save_workbook()
        """
        if self.workbook:
            self.close_workbook()

        fmt = str(fmt).lower().strip()
        if fmt == "xlsx":
            self.workbook = XlsxWorkbook(path)
        elif fmt == "xls":
            self.workbook = XlsWorkbook(path)
        else:
            raise ValueError(f"Unknown format: {fmt}")

        self.workbook.create()
        if sheet_name is not None:
            self.rename_worksheet(self.get_active_worksheet(), sheet_name)

        return self.workbook

    def open_workbook(
        self,
        path: str,
        data_only: Optional[bool] = False,
        read_only: Optional[bool] = False,
    ) -> Union["XlsWorkbook", "XlsxWorkbook"]:
        """Open an existing Excel workbook.

        Opens the workbook in memory and sets it as the active workbook.
        **This library can only have one workbook open at a time, and
        any previously opened workbooks are closed first.**

        The file can be in either ``.xlsx`` or ``.xls`` format.

        :param path: path to Excel file
        :param data_only: controls whether cells with formulas have either
         the formula (default, False) or the value stored the last time Excel
         read the sheet (True). Affects only ``.xlsx`` files.
        :return: Workbook object

        Examples:

        .. code-block:: robotframework

            # Open workbook with only path provided
            Open Workbook    path/to/file.xlsx

            # Open workbook with path provided and reading formulas in cells
            # as the value stored
            # Note: Can only be used with XLSX workbooks
            Open Workbook    path/to/file.xlsx    data_only=True

        .. code-block:: python

            # Open workbook with only path provided
            lib.open_workbook(path="path/to/file.xlsx")

            # Open workbook with path provided and reading formulas in cells
            # as the value stored
            # Note: Can only be used with XLSX workbooks
            lib.open_workbook(path="path/to/file.xlsx", data_only=True)
        """
        if self.workbook:
            self.close_workbook()

        self.workbook = self._load_workbook(path, data_only, read_only)
        self.logger.info("Opened workbook: %s", self.workbook)
        return self.workbook

    def close_workbook(self) -> None:
        """Close the active workbook.

        Examples:

        .. code-block:: robotframework

            # Close active workbook
            Close Workbook

        .. code-block:: python

            # Close active workbook
            lib.close_workbook()
        """

        if self.workbook:
            self.logger.info("Closing workbook: %s", self.workbook)
            self.workbook.close()
            self.workbook = None

    def save_workbook(
        self, path: Optional[str] = None
    ) -> Union["XlsWorkbook", "XlsxWorkbook"]:
        """Save the active workbook.

        **Note:** No changes to the workbook are saved to the actual file unless
        this keyword is called.

        :param path: Path to save to. If not given, uses path given
                     when opened or created.
        :return:     Workbook object

        Examples:

        .. code-block:: robotframework

            # Saving the active workbook to a new location/filename or saving to
            # a new location/filename
            # Note: You cannot use Save Workbook to convert from XLSX to XLS
            # or vice-versa
            Save Workbook   path=${OUTPUT_DIR}${/}orders.xlsx

            # Saving the active workbook changes if location/filename were set
            # in Create Workbook or Open Workbook
            Save Workbook

        .. code-block:: python

            # Saving the active workbook to a new location/filename or saving to
            # a new location/filename
            # Note: You cannot use Save Workbook to convert from XLSX to XLS
            # or vice-versa
            lib.save_workbook(path="./output/orders.xlsx")

            # Saving the active workbook changes if location/filename were set
            # in Create Workbook or Open Workbook
            lib.save_workbook()

        """
        assert self.workbook, "No active workbook"

        try:
            extension = pathlib.Path(path).suffix
        except TypeError:
            extension = None

        if (
            self.workbook.extension is not None
            and extension is not None
            and self.workbook.extension != extension
        ):
            self.logger.warning(
                "Changed file extension from %s to %s",
                self.workbook.extension,
                extension,
            )

        self.workbook.validate_content()
        return self.workbook.save(path)

    def list_worksheets(self) -> List[str]:
        """List all names of worksheets in the given workbook.

        :return:    List containing the names of the worksheets

        Examples:

        .. code-block:: robotframework

            # List Worksheets will read the worksheet names into a list variable
            # The variable should be declared with the List type "@" when being used
            # to store the sheet names from the List Worksets keyword
            @{sheets}=    List Worksheets

        .. code-block:: python

            # List Worksheets will read the worksheet names into a list variable
            # The variable should be declared with the List type "@" when being used
            # to store the sheet names from the List Worksets keyword
            sheets = lib.list_worksheets()

        """
        assert self.workbook, "No active workbook"
        return self.workbook.sheetnames

    def worksheet_exists(self, name: str) -> bool:
        """Return True if worksheet with given name is in workbook.

        :param name: Name of worksheet you are looking for
        :return: `True` if the worksheet exists, `False` otherwise

        Examples:

        .. code-block:: robotframework

            # To use Worksheet Exists in a conditional statement set it to
            # a variable first, like you see here, and then compare the
            # variable to ${TRUE} or ${FALSE}
            ${Does_Worksheet_Exist}=    Worksheet Exists    Sheet

        .. code-block:: python

            Does_Worksheet_Exist = lib.worksheet_exists("Sheet")
        """
        assert self.workbook, "No active workbook"
        return bool(str(name) in self.list_worksheets())

    def get_active_worksheet(self) -> str:
        """Get the name of the worksheet which is currently active.


        :return:    Active worksheet name

        Examples:

        .. code-block:: robotframework

            ${Active_Worksheet}=    Get Active Worksheet

        .. code-block:: python

            Active_Worksheet = lib.get_active_worksheet()
        """

        assert self.workbook, "No active workbook"
        return self.workbook.active

    def set_active_worksheet(self, value: Union[str, int]) -> None:
        """Set the active worksheet.

        This keyword can be used to set the default worksheet for keywords,
        which removes the need to specify the worksheet name for each keyword.
        It can always be overridden on a per-keyword basis.

        :param value: Index or name of worksheet

        Examples:

        .. code-block:: robotframework

            # Set using the name of the worksheet
            Set Active Worksheet    Customers

            # Set using the index of the worksheet
            # Worksheet index begings at 0
            Set Active Worksheet    2

        .. code-block:: python

            # Set using the name of the worksheet
            lib.set_active_worksheet("Customers")

            # Set using the index of the worksheet
            # Worksheet index begings at 0
            lib.set_active_worksheet(2)
        """
        assert self.workbook, "No active workbook"
        self.workbook.active = value

    def create_worksheet(
        self,
        name: str,
        content: Optional[Any] = None,
        exist_ok: Optional[bool] = False,
        header: Optional[bool] = False,
    ) -> None:
        """Create a new worksheet in the current workbook.

        :param name:     Name of new worksheet
        :param content:  Optional content for worksheet
        :param exist_ok: If `False`, raise an error if name is already in use
        :param header:   If content is provided, write headers to worksheet

        Examples:

        .. code-block:: robotframework

            # Create a new blank worksheet named "Customers"
            Create Worksheet    Customers

            # Create a new worksheet with headers and contents using
            # a List of Dictonaries
            # Don't forget to `Save Workbook` once your changes are complete
            &{Employees_Row1}=    Create Dictionary    name=Mark    age=${58}
            &{Employees_Row2}=    Create Dictionary    name=John    age=${22}
            &{Employees_Row3}=    Create Dictionary    name=Adam    age=${67}
            @{Worksheet_Data}=    Create List
            ...    ${Worksheet_Data_row1}
            ...    ${Worksheet_Data_row2}
            ...    ${Worksheet_Data_row3}
            Create Worksheet
            ...    name=Employees
            ...    content=${Worksheet_Data}
            ...    header=True
            Save Workbook

            # Create a new workseet using a Dictionary of Lists
            # Don't forget to `Save Workbook` once your changes are complete
            @{Employees_name}=    Create List    Mark    John    Adam
            @{Employees_age}=    Create List    ${58}    ${22}    ${67}
            &{Worksheet_Data}=    Create Dictionary
            ...    name=${Worksheet_Data_name}
            ...    age=${Worksheet_Data_age}
            Create Worksheet
            ...    name=Employees
            ...    content=${Worksheet_Data}
            ...    header=True
            Save Workbook

        .. code-block:: python

            # Create a new blank worksheet named "Customers"
            lib.create_worksheet("Customers")

            # Create a new workseet using a List of Dictionaries
            # Don't forget to `Save Workbook` once your changes are complete
            Worksheet_Data = [
                {"name": "Mark", "age": 58},
                {"name": "John", "age": 22},
                {"name": "Adam", "age": 67},
                ]
            lib.create_worksheet(name="Employees",content=Worksheet_Data,header=True)
            lib.save_workbook()

            # Create a new workseet using a Dictionary of Lists
            # Don't forget to `Save Workbook` once your changes are complete
            Worksheet_Data = {
                "name": ["Mark", "John", "Adam"],
                "age":  [    58,     22,     67],
                }
            lib.create_worksheet(name="Employees",content=Worksheet_Data,header=True)
            lib.save_workbook()
        """
        assert self.workbook, "No active workbook"
        if name in self.workbook.sheetnames and not exist_ok:
            raise ValueError(f"Sheet with name {name} already exists")

        self.workbook.create_worksheet(name)
        if content:
            self.workbook.append_worksheet(name, content, header)

    def read_worksheet(
        self,
        name: Optional[str] = None,
        header: Optional[bool] = False,
        start: Optional[int] = None,
    ) -> List[dict]:
        """Read the content of a worksheet into a list of dictionaries.

        Each key in the dictionary will be either values from the header row,
        or Excel-style column letters.

        :param name:   Name of worksheet to read (optional).
                       Defaults to the active worksheet.
        :param header: If `True`, use the first row of the worksheet
                       as headers for the rest of the rows. Default is `False`.
        :param start:  Row index to start reading data from (1-indexed).
                       Default value is row 1.
        :return:       List of dictionaries that represents the worksheet

        Examples:

        .. code-block:: robotframework

            # The most simple form. Column keys will be Column letters.
            ${rows}=    Read Worksheet

            # Since `header=True` the keys will be the header values
            ${rows}=    Read Worksheet     header=True

            # Uses the header values as keys and starts reading at row 3
            ${rows}=    Read Worksheet     header=True    start=${3}

        .. code-block:: python

            # The most simple form. Keys will be Column letters.
            rows = lib.read_worksheet()

            # Since `header=True` the keys will be the header values
            rows = lib.read_worksheet(header=True)

            # Uses the header values as keys and starts reading at row 3
            rows = lib.read_worksheet(header=True, start=3)
        """
        assert self.workbook, "No active workbook"
        return self.workbook.read_worksheet(name, header, start)

    def read_worksheet_as_table(
        self,
        name: Optional[str] = None,
        header: bool = False,
        trim: bool = True,
        start: Optional[int] = None,
    ) -> Table:
        """Read the contents of a worksheet into a Table container. Allows
        sorting/filtering/manipulating using the ``RPA.Tables`` library.

        :param name:   Name of worksheet to read (optional).
                       Defaults to the active worksheet.
        :param header: If `True`, use the first row of the worksheet
                       as headers for the rest of the rows. Default value is False.
        :param trim:   Remove all empty rows from the end of the worksheet.
                       Default value is True.
        :param start:  Row index to start reading data from (1-indexed).
                       Default value is row 1.
        :return:       Table object that represents the worksheet

        Examples:

        .. code-block:: robotframework

            # The most simple form. Column keys will be Column letters.
            ${table}=    Read Worksheet As Table

            # Since `header=True` the keys will be the header values
            ${table}=    Read Worksheet As Table     header=True

            # Uses the header values as keys and starts reading at row 3
            ${table}=    Read Worksheet As Table     header=True    start=${3}

        .. code-block:: python

            # The most simple form. Keys will be Column letters.
            table = lib.read_worksheet_as_table()

            # Since `header=True` the keys will be the header values
            table = lib.read_worksheet_as_table(header=True)

            # Uses the header values as keys and starts reading at row 3
            table = lib.read_worksheet_as_table(header=True, start=3)
        """
        tables = Tables()
        sheet = self.read_worksheet(name, header, start)
        return tables.create_table(sheet, trim)

    def append_rows_to_worksheet(
        self,
        content: Any,
        name: Optional[str] = None,
        header: bool = False,
        start: Optional[int] = None,
        formatting_as_empty: Optional[bool] = False,
    ) -> List[dict]:
        """Append values to the end of the worksheet.

        :param content: Rows of values to append
        :param name:    Name of worksheet to append to (optional).
                        Defaults to the active worksheet.
        :param header:  Set rows according to existing header row
        :param start:   Start of data, NOTE: Only required when header is True
        :param formatting_as_empty: if True, the cells containing only
                        formatting (no values) are considered empty.
        :return:        List of dictionaries that represents the worksheet

        The ``content`` argument can be of any tabular format. Typically,
        this is a Table object created by the ``RPA.Tables`` library,
        but it can also be a list of lists, or a list of dictionaries.

        If the ``header`` flag is enabled, the existing header in the worksheet
        is used to insert values in the correct columns. This assumes that
        that source data has this data available.

        If the header is not on the first row of the worksheet,
        the ``start`` argument can be used to give the correct row index.

        Examples:

        .. code-block:: robotframework

            # Append an existing Table object
             # Create a new table using a Dictionary of Lists
            @{table_name}=    Create List    Sara    Beth    Amy
            @{table_age}=    Create List    ${48}    ${21}    ${57}
            &{table}=    Create Dictionary    name=${table_name}    age=${table_age}
            Create Table    ${table}
            Append rows to worksheet    ${table}
            Save Workbook

            # Append to a worksheet with headers on row 5
             # Create a new table using a Dictionary of Lists
            @{table_name}=    Create List    Sara    Beth    Amy
            @{table_age}=    Create List    ${48}    ${21}    ${57}
            &{table}=    Create Dictionary    name=${table_name}    age=${table_age}
            Create Table    ${table}
            Append rows to worksheet    ${table}    header=${TRUE}   start=5
            Save Workbook

        .. code-block:: python

            # Append an existing Table object
            table = {
                "name": ["Sara", "Beth", "Amy"],
                "age":  [    48,     21,     57],
                }
            lib.append_rows_to_worksheet(table)
            lib.save_workbook()

            # Append to a worksheet with headers on row 5
            table = {
                "name": ["Sara", "Beth", "Amy"],
                "age":  [    48,     21,     57],
                }
            lib.append_rows_to_worksheet(table, header=True, start=5)
            lib.save_workbook()
        """
        assert self.workbook, "No active workbook"
        return self.workbook.append_worksheet(
            name, content, header, start, formatting_as_empty
        )

    def remove_worksheet(self, name: str = None) -> None:
        """Remove a worksheet from the active workbook.

        :param name: Name of worksheet to remove (optional).
                     Defaults to the active worksheet.

        Examples:

        .. code-block:: robotframework

            # Remove last worksheet
            ${sheets}=       List worksheets
            Remove worksheet    ${sheets}[-1]

            # Remove worksheet by name
            Remove Worksheet    Sheet

        .. code-block:: python

            # Remove last worksheet
            sheets = lib.list_worksheets()
            lib.remove_worksheet(sheets[-1])

            # Remove worksheet by name
            lib.remove_worksheet("Sheet")
        """
        assert self.workbook, "No active workbook"
        self.workbook.remove_worksheet(name)

    def rename_worksheet(self, src_name: str, dst_name: str) -> None:
        """Rename an existing worksheet in the active workbook.

        :param src_name: Current name of worksheet
        :param dst_name: Future name of worksheet

        Examples:

        .. code-block:: robotframework

            Rename worksheet    Sheet    Orders

        .. code-block:: python

            lib.rename_worksheet("Sheet","Orders")
        """
        assert self.workbook, "No active workbook"
        self.workbook.rename_worksheet(dst_name, src_name)

    def find_empty_row(self, name: Optional[str] = None) -> int:
        """Find the first empty row after existing content,
        and return the row number.

        :param name:    Name of worksheet (optional). Defaults to the active worksheet.
        :return:        First row number of empty row

        Examples:

        .. code-block:: robotframework

            ${next}=    Find empty row

        .. code-block:: python

            next = lib.find_empty_row()
        """
        assert self.workbook, "No active workbook"
        return self.workbook.find_empty_row(name)

    def get_cell_value(
        self, row: int, column: Union[str, int], name: Optional[str] = None
    ) -> Any:
        """Get a cell value in the given worksheet.

        :param row:     Index of row to read, e.g. 3
        :param column:  Name or index of column, e.g. C or 7
        :param name:    Name of worksheet (optional). Defaults to active worksheet.
        :return:        Cell value

        Examples:

        .. code-block:: robotframework

            # Read header names
            ${column1}=    Get cell value    1    A
            ${column2}=    Get cell value    1    B
            ${column3}=    Get cell value    1    C

        .. code-block:: python

            # Read header names
            column1 = lib.get_cell_value(1, "A")
            column2 = lib.get_cell_value(1, "B")
            column3 = lib.get_cell_value(1, "C")
        """
        assert self.workbook, "No active workbook"
        return self.workbook.get_cell_value(row, column, name)

    def set_cell_value(
        self,
        row: int,
        column: Union[str, int],
        value: Any,
        name: Optional[str] = None,
        fmt: Optional[Union[str, float]] = None,
    ) -> None:
        """Set a cell value in the given worksheet.

        :param row:     Index of row to write, e.g. 3
        :param column:  Name or index of column, e.g. C or 7
        :param value:   New value of cell
        :param name:    Name of worksheet (optional). Defaults to active worksheet.
        :param fmt:     Format code for cell (optional)

        Examples:

        .. code-block:: robotframework

            # Set a value in the first row and column
            Set cell value    1    1    Some value
            Set cell value    1    A    Some value

            # Set a value with cell formatting
            Set cell value    2    B    ${value}    fmt=0%

        .. code-block:: python

            # Set a value in the first row and column
            lib.set_cell_value(1, 1, "Some value")
            lib.set_cell_value(1, "A", "Some value")

            # Set a value with cell formatting
            lib.set_cell_value(2, "B", value, fmt="0%")
        """
        assert self.workbook, "No active workbook"

        self.workbook.set_cell_value(row, column, value, name)

        if fmt is not None:
            self.workbook.set_cell_format(row, column, fmt, name)

    def set_cell_format(
        self,
        row: int,
        column: Union[str, int],
        fmt: Union[str, float],
        name: Optional[str] = None,
    ) -> None:
        """Set format for cell.

        Does not affect the values themselves, but changes how the values
        are displayed when opening with an external application such as
        Microsoft Excel or LibreOffice Calc.

        :param row:     Index of row to write, e.g. 3
        :param column:  Name or index of column, e.g. C or 7
        :param fmt:     Format code for cell
        :param name:    Name of worksheet (optional). Defaults to active worksheet.

        The ``fmt`` argument accepts all format code values that
        are supported by the aforementioned applications.

        Some examples of valid values:

        ======== ===========
        Format   Explanation
        ======== ===========
        0.00     Number with two decimal precision
        0%       Percentage without decimals
        MM/DD/YY Date with month, day, and year
        @        Text value
        BOOLEAN  Boolean value
        ======== ===========

        Examples:

        .. code-block:: robotframework

            # Set value to have one decimal precision
            Set cell format   2  B    00.0

        .. code-block:: python

            # Set value to have one decimal precision
            lib.set_cell_format(2, "B", 00.0)
        """
        assert self.workbook, "No active workbook"
        self.workbook.set_cell_format(row, column, fmt, name)

    def insert_image_to_worksheet(
        self,
        row: int,
        column: Union[int, str],
        path: str,
        scale: float = 1.0,
        name: Optional[str] = None,
    ) -> None:
        """Insert an image into the given cell.

        The ``path`` argument should be a local file path to the image file.

        By default the image is inserted in the original size, but it can
        be scaled with the ``scale`` argument. It's scaled with a factor
        where the value ``1.0`` is the default.

        :param row:     Index of row to write
        :param column:  Name or index of column
        :param path:    Path to image file
        :param scale:   Scale of image (optional). Default value is "1.0".
        :param name:    Name of worksheet (optional). Defaults to the active worksheet.

        Examples:

        .. code-block:: robotframework

            Insert image to worksheet    ${last_row}    A    screenshot.png

        .. code-block:: python

            lib.insert_image_to_worksheet(last_row, "A", "screenshot.png")
        """
        assert self.workbook, "No active workbook"
        image = Image.open(path)

        if scale != 1.0:
            fmt = image.format
            width = int(image.width * float(scale))
            height = int(image.height * float(scale))
            image = image.resize((width, height), Image.Resampling.LANCZOS)
            image.format = fmt

        self.workbook.insert_image(row, column, image, name)

    def get_worksheet_value(
        self, row: int, column: Union[str, int], name: Optional[str] = None
    ) -> Any:
        """Alias for keyword ``Get cell value``, see the original keyword
        for documentation.
        """
        # Old keyword name, deprecate at some point
        return self.get_cell_value(row, column, name)

    def set_worksheet_value(
        self,
        row: int,
        column: Union[str, int],
        value: Any,
        name: Optional[str] = None,
        fmt: Optional[Union[str, float]] = None,
    ) -> Any:
        """Alias for keyword ``Set cell value``, see the original keyword
        for documentation.
        """
        # Old keyword name, deprecate at some point
        return self.set_cell_value(row, column, value, name, fmt)

    def clear_cell_range(
        self,
        range_string: str,
    ):
        """Clear cell values for a given range.

        :param range_string: single cell or range of cells

        Examples:

        Robot Framework example.

        .. code-block:: robotframework

            # area of cells
            Clear Cell Range    A9:A100
            # single cell
            Clear Cell Range    A2

        Python example.

        .. code-block:: python

            lib.clear_cell_range("A1")
            lib.clear_cell_range("B2:B50")
        """
        self._require_open_xlsx_workbook("clear_cell_range")
        cr = CellRange(range_string=range_string)
        for row, acell in list(cr.cells):
            self.workbook.book.active.cell(row, acell).value = None

    def delete_rows(self, start: int, end: Optional[int] = None):
        """Delete row or rows beginning from start row number to
        possible end row number.

        :param start: row number to start deletion from
        :param end: optional row number for last row to delete

        Examples:

        Robot Framework example.

        .. code-block:: robotframework

            Delete Rows   2       # delete row 2
            Delete Rows   5  10   # delete rows 5-10

        Python example.

        .. code-block:: python

            lib.delete_rows(2)
            lib.delete_rows(5,10)
        """
        self._require_open_xlsx_workbook("delete_rows")
        amount = (end - start + 1) if end else 1
        self.workbook.book.active.delete_rows(start, amount)

    def delete_columns(
        self, start: Union[int, str], end: Optional[Union[int, str]] = None
    ):
        """Delete column or columns beginning from start column number/name to
        possible end column number/name.

        :param start: column number or name to start deletion from
        :param end: optional column number or name for last column to delete

        Examples:

        Robot Framework example.

        .. code-block:: robotframework

            Delete Columns   C       # delete column C
            Delete Columns   3       # delete column 3 (same as C)
            Delete Columns   E  AA   # delete rows E-AA

        Python example.

        .. code-block:: python

            lib.delete_columns("D")
            lib.delete_rows(1, "JJ")
        """
        self._require_open_xlsx_workbook("delete_columns")
        start_column_index = (
            start if isinstance(start, int) else column_index_from_string(start)
        )
        amount = 1
        if end:
            end_column_index = (
                end if isinstance(end, int) else column_index_from_string(end)
            )
            amount = end_column_index - start_column_index + 1
        self.workbook.book.active.delete_cols(start_column_index, amount)

    def insert_columns_before(self, column: Union[int, str], amount: int = 1):
        """Insert column or columns before a column number/name.

        :param column: insert before this column
        :param amount: number of columns to insert, default 1

        Examples:

        Robot Framework example.

        .. code-block:: robotframework

            Insert Columns Before   C      # insert 1 column before column C
            Insert Columns Before   A  3   # insert 3 columns before column A

        Python example.

        .. code-block:: python

            lib.insert_columns_before("C")
            lib.insert_columns_before("A", 3)
        """
        self._require_open_xlsx_workbook("insert_columns_before")

        column_index = (
            column_index_from_string(column) if isinstance(column, str) else column
        )
        self.workbook.book.active.insert_cols(column_index, amount)

    def insert_columns_after(self, column: Union[int, str], amount: int = 1):
        """Insert column or columns after a column number/name.

        :param column: insert after this column
        :param amount: number of columns to insert, default 1

        Examples:

        Robot Framework example.

        .. code-block:: robotframework

            Insert Columns After   C      # insert 1 column after column C
            Insert Columns Before   A  3   # insert 3 columns after column A

        Python example.

        .. code-block:: python

            lib.insert_columns_after("C")
            lib.insert_columns_after("A", 3)
        """
        self._require_open_xlsx_workbook("insert_columns_after")

        column_index = (
            column_index_from_string(column) if isinstance(column, str) else column
        )
        self.workbook.book.active.insert_cols(column_index + amount - 1, amount)

    def insert_rows_before(self, row: int, amount: int = 1):
        """Insert row or rows before a row number.

        :param row: insert before this row
        :param amount: number of rows to insert, default 1

        Examples:

        Robot Framework example.

        .. code-block:: robotframework

            Insert Rows Before   3      # insert 1 row before row 3
            Insert Rows Before   1  3   # insert 3 rows before row 1

        Python example.

        .. code-block:: python

            lib.insert_rows_before(1)
            lib.insert_rows_before(1, 3)
        """
        self._require_open_xlsx_workbook("insert_rows_before")

        self.workbook.book.active.insert_rows(row, amount)

    def insert_rows_after(self, row: int, amount: int = 1):
        """Insert row or rows after a row number.

        :param row: insert after this row
        :param amount: number of rows to insert, default 1

        Examples:

        Robot Framework example.

        .. code-block:: robotframework

            Insert Rows After   3      # insert 1 row after row 3
            Insert Rows After   1  3   # insert 3 rows after row 1

        Python example.

        .. code-block:: python

            lib.insert_rows_after(1)
            lib.insert_rows_after(1, 3)
        """
        self._require_open_xlsx_workbook("insert_rows_after")

        self.workbook.book.active.insert_rows(row + amount - 1, amount)

    def copy_cell_values(self, source_range: str, target: str):
        """Copy cells from source to target.

        :param source_range: single cell or range of cells
        :param target: copy to this cell

        Examples:

        Robot Framework example.

        .. code-block:: robotframework

            Copy Cell Values   A1:D4   G10

        Python example.

        .. code-block:: python

            lib.copy_cell_values("A1:D4", "G10")
        """
        self._require_open_xlsx_workbook("copy_cell_values")

        cr = CellRange(range_string=source_range)
        cells = list(cr.cells)
        target_cell_unpacked = utils_cell.coordinate_from_string(target)
        target_column = column_index_from_string(target_cell_unpacked[0])
        target_row = target_cell_unpacked[1]

        last_row = None
        column_index = 0
        for row, column in cells:
            if not last_row:
                last_row = row
            if last_row != row:
                target_row += 1
                last_row = row
                column_index = 0
            target_cell = self.workbook.book.active.cell(
                target_row, target_column + column_index
            )
            source_cell = self.workbook.book.active.cell(row, column)
            target_cell.value = source_cell.value
            column_index += 1

    def set_styles(
        self,
        range_string: str,
        font_name: str = None,
        family: str = None,
        size: int = None,
        bold: bool = False,
        italic: bool = False,
        underline: bool = False,
        strikethrough: bool = False,
        cell_fill: str = None,
        color: str = None,
        align_horizontal: str = None,
        align_vertical: str = None,
        number_format: str = None,
    ):
        """Set styles for range of cells.

        Possible values for the `align_horizontal`:

            - general
            - left
            - center
            - right
            - fill
            - justify
            - centerContinuous
            - distributed

        Possible values for the `align_vertical`:

            - top
            - center
            - bottom
            - justify
            - distributed

        Some examples for `number_formats`:

            - General
            - 0
            - 0.00
            - #,##0
            - #,##0.00
            - "$"#,##0_);("$"#,##0)
            - "$"#,##0_);[Red]("$"#,##0)
            - 0%
            - 0.00%
            - 0.00E+00
            - # ?/?
            - # ??/??
            - mm-dd-yy
            - d-mmm-yy
            - d-mmm
            - h:mm AM/PM
            - h:mm:ss AM/PM
            - h:mm
            - h:mm:ss
            - m/d/yy h:mm

        :param range_string: single cell or range of cells
        :param font_name: name of the font
        :param family: font family name
        :param size: size for the font
        :param bold: font style bold
        :param italic: font style italics
        :param underline: font style underline
        :param strikethrough: font style strikethrough
        :param cell_fill: cell fill color, in hex or color name
        :param color: font color, in hex or color name
        :param align_horizontal: cell horizontal alignment
        :param align_vertical: cell vertical alignment
        :param number_format: cell number format

        Examples:

        Robot Framework example.

        .. code-block:: robotframework

            Set Styles    A1:D4
            ...  bold=True
            ...  cell_fill=lightblue
            ...  align_horizontal=center
            ...  number_format=h:mm AM/PM

            Set Styles    E2
            ...  strikethrough=True
            ...  color=FF0000

        Python example.

        .. code-block:: python

            lib.set_styles("A1:D4", bold=True, font_name="Arial", size=24)
        """
        self._require_open_xlsx_workbook("set_styles")

        cr = CellRange(range_string=range_string)
        font_parameters = {}
        self._set_font_param_if_given(font_parameters, "name", font_name)
        self._set_font_param_if_given(font_parameters, "family", family)
        self._set_font_param_if_given(font_parameters, "sz", size)
        self._set_font_param_if_given(font_parameters, "b", bold)
        self._set_font_param_if_given(font_parameters, "i", italic)
        self._set_font_param_if_given(font_parameters, "u", underline)
        self._set_font_param_if_given(font_parameters, "strike", strikethrough)
        self._set_color_if_given(font_parameters, color)

        for row, column in list(cr.cells):
            active_cell = self.workbook.book.active.cell(row, column)
            active_cell.font = Font(**font_parameters)
            self._set_fill_color(active_cell, cell_fill)
            self._set_cell_alignments(active_cell, align_horizontal, align_vertical)
            self._set_cell_number_format(active_cell, number_format)

    def _set_font_param_if_given(self, parameters, param_name, value):
        if value:
            parameters[param_name] = value

    def _set_color_if_given(self, parameters, value):
        if value:
            match = re.search(r"^(?:[0-9a-fA-F]{3}){2}$", value)
            if match:
                color_hex = value
            else:
                color = ImageColor.getrgb(value)
                color_hex = "%02x%02x%02x" % color
            parameters["color"] = xlsColor(color_hex)

    def _set_fill_color(self, active_cell, value):
        if value:
            match = re.search(r"^(?:[0-9a-fA-F]{3}){2}$", value)
            if match:
                color_hex = value
            else:
                color = ImageColor.getrgb(value)
                color_hex = "%02x%02x%02x" % color
            active_cell.fill = PatternFill("solid", color_hex, color_hex)

    def _set_cell_alignments(self, active_cell, horizontal, vertical):
        if horizontal or vertical:
            active_cell.alignment = Alignment(horizontal=horizontal, vertical=vertical)

    def _set_cell_number_format(self, active_cell, number_format):
        if number_format:
            active_cell.number_format = number_format

    def auto_size_columns(
        self,
        start_column: Union[int, str],
        end_column: Optional[Union[int, str]] = None,
        width: Optional[int] = None,
    ):
        """Auto size column widths.

        Note. non-default font sizes might cause auto sizing issues

        :param start_column: column number or name to start from
        :param end_column: optional column number or name for last column
        :param width: if given will resize columns to this size, otherwise
         will auto_size

        Examples:

        Robot Framework example.

        .. code-block:: robotframework

            Auto Size Columns   A   D    # will try auto size
            Auto Size Columns   B   D   16  # will set A-D columns sizes to 16
            Auto Size Columns   A   width=24  # will set column A size to 24

        Python example.

        .. code-block:: python

            lib.auto_size_columns("A", "D")
            lib.auto_size_columns("C", width=40)
        """
        self._require_open_xlsx_workbook("auto_size_columns")
        start_index = (
            column_index_from_string(start_column)
            if isinstance(start_column, str)
            else start_column
        )
        end_index = start_index
        if end_column:
            end_index = (
                column_index_from_string(end_column)
                if isinstance(end_column, str)
                else end_column
            )

        for col in range(start_index, end_index + 1):
            col_letter = get_column_letter(col)
            if width:
                self.workbook.book.active.column_dimensions[col_letter].width = width
            else:
                self.workbook.book.active.column_dimensions[col_letter].auto_size = True

    def hide_columns(
        self,
        start_column: Union[int, str],
        end_column: Optional[Union[int, str]] = None,
    ):
        """Hide column or columns in worksheet.

        :param start_column: column number or name to start from
        :param end_column: optional column number or name for last column

        Examples:

        Robot Framework example.

        .. code-block:: robotframework

            Hide Columns   A   D    # hide columns A-D
            Hide Columns   A        # hide column A

        Python example.

        .. code-block:: python

            lib.hide_columns("A", "D")
            lib.hide_columns("A")
        """
        self._require_open_xlsx_workbook("hide_columns")

        self._set_column_hidden(True, start_column, end_column)

    def unhide_columns(
        self,
        start_column: Union[int, str],
        end_column: Optional[Union[int, str]] = None,
    ):
        """Unhide column or columns in worksheet.

        :param start_column: column number or name to start from
        :param end_column: optional column number or name for last column

        Examples:

        Robot Framework example.

        .. code-block:: robotframework

            Unhide Columns   A   D    # unhide columns A-D
            Unhide Columns   A        # unhide column A

        Python example.

        .. code-block:: python

            lib.unhide_columns("A", "D")
            lib.unhide_columns("A")
        """
        self._require_open_xlsx_workbook("unhide_columns")

        self._set_column_hidden(False, start_column, end_column)

    def _set_column_hidden(self, hidden: bool, start_column, end_column):
        start_index = (
            column_index_from_string(start_column)
            if isinstance(start_column, str)
            else start_column
        )
        end_index = start_index
        if end_column:
            end_index = (
                column_index_from_string(end_column)
                if isinstance(end_column, str)
                else end_column
            )

        for col in range(start_index, end_index + 1):
            col_letter = get_column_letter(col)
            self.workbook.book.active.column_dimensions[col_letter].hidden = hidden

    def set_cell_formula(
        self, range_string: str, formula: str, transpose: bool = False
    ):
        """Set cell formula for given range of cells.

        If `transpose` is set then formula is set for first cell of the
        range and the rest of cells will transpose the function to match
        to that cell.

        Otherwise (by default) all cells will get the same formula.

        :param range_string: cell range
        :param formula: formula for the cell
        :param transpose: on True the cell formulas will be transposed

        Examples:

        Robot Framework example.

        .. code-block:: robotframework

            # all cells will have same formula
            Set Cell Formula   E2:E10    =B2+5
            # cells will have transposed formulas
            # E2 will have =B2+5
            # E3 will have =B3+5
            # etc
            Set Cell Formula   E2:E10    =B2+5   True

        Python example.

        .. code-block:: python

            lib.set_cell_formula("E2:E10", "=B2+5")
            lib.set_cell_formula("E2:E10", "=B2+5", True)
        """
        self._require_open_xlsx_workbook("set_cell_formula")

        cr = CellRange(range_string=range_string)
        start_col, start_row, _, _ = cr.bounds
        start_col_str = f"{get_column_letter(start_col)}{start_row}"
        cells = list(cr.cells)

        for index, acell in enumerate(cells):
            row, column = acell
            col_str = f"{get_column_letter(column)}{row}"
            if (transpose and index == 0) or not transpose:
                self.workbook.book.active[col_str].value = formula
            elif transpose and index > 0:
                self.workbook.book.active[col_str] = Translator(
                    formula, origin=start_col_str
                ).translate_formula(col_str)

    def move_range(
        self, range_string: str, rows: int = 0, columns: int = 0, translate: bool = True
    ):
        """Move range of cells by given amount of rows and columns.

        Formulas are translated to match new location by default.

        *Note*. There is a bug in the openpyxl on moving negative rows/columns.

        :param range_string: cell range
        :param rows: number of rows to move
        :param columns: number of columns to move
        :param translate: are formulas translated for a new location

        Examples:

        Robot Framework example.

        .. code-block:: robotframework

            # move range 4 rows down
            Move Range   E2:E10    rows=4
            # move range 2 rows down, 2 columns right
            Move Range   E2:E10    rows=2  columns=2

        Python example.

        .. code-block:: python

            lib.move_range("E2:E10", rows=4)
            lib.move_range("E2:E10", rows=2, columns=2)
        """
        self._require_open_xlsx_workbook("move_range")

        self.workbook.book.active.move_range(
            range_string, rows=rows, cols=columns, translate=translate
        )

    def set_cell_values(
        self, start_cell: str, values: Union[list, Table], table_heading: bool = False
    ):
        """Set cell values given as list of lists or as a `RPA.Tables.Table`.

        *Note.* Will overwrite cells if table structure causes cells to overlap.

        :param start_cell: starting cell in a string
        :param values: list of lists or a Table
        :param table_heading: if values are given as a Table, this parameter
         defines if Table headings should be inserted as a row

        Examples:

        Robot Framework example.

        .. code-block:: robotframework

            @{all_rows}=    Create List
            ${headers}=    Create List    first    second   third  fourth
            FOR    ${num}    IN RANGE    1    2000
                @{row}=    Create List    ${num}    ${num+1}    ${num*2}    ${num*4}
                Append To List    ${all_rows}    ${row}
            END
            #  Set Cell Values from Table (include headers)
            ${table}=    Create Table    ${all_rows}    columns=${headers}
            Set Cell Values   G1   ${table}   True
            #  Set Cell Values from a list of lists
            # uncomment if headings should be added
            # Append To List  ${all_rows}   ${headers}
            Set Cell Values   M1   ${all_rows}

            # Simplest form of adding values
            @{values}=    Evaluate    [[1,2,3],[4,5,6],['a','b','c','d']]
            Set Cell Values   A1   ${values}

        Python example.

        .. code-block:: python

            data =  [[1,2,3],[4,5,6],['a','b','c','d']]
            lib.set_cell_values("E2", data)
        """
        self._require_open_xlsx_workbook("set_cell_values")
        if isinstance(values, Table):
            data_values = return_table_as_raw_list(values, table_heading)
        else:
            data_values = values
        cr = CellRange(range_string=start_cell)
        start_col, start_row, _, _ = cr.bounds
        for row_index, row in enumerate(data_values):
            if isinstance(row, list):
                for col_index, col in enumerate(row):
                    column = start_col + col_index
                    row = start_row + row_index
                    self.workbook.book.active[
                        f"{get_column_letter(column)}{row}"
                    ].value = col
            else:
                column = start_col + row_index
                row = start_row
                self.workbook.book.active[f"{get_column_letter(column)}{row}"].value = (
                    row
                )


class BaseWorkbook:
    """Common logic for both .xls and .xlsx files management."""

    def __init__(self, path: Optional[PathType] = None):
        self.logger = logging.getLogger(__name__)
        self.path = path
        self._book = None
        self._extension = None
        self._active = None

    @property
    def book(self):
        return self._book

    def _validate_content(self, props_obj: Any):
        # Strips leading/trailing whitespace in Excel properties.
        public_props = [prop for prop in dir(props_obj) if not prop.startswith("_")]
        for prop in public_props:
            value = getattr(props_obj, prop)
            if value and isinstance(value, str):
                setattr(props_obj, prop, value.strip())


class XlsxWorkbook(BaseWorkbook):
    """Container for manipulating modern Excel files (.xlsx)"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._read_only = False

    @staticmethod
    def is_sheet_empty(sheet):
        # Maximum rows/columns are always 1 or more, even when the sheet doesn't
        #  contain cells at all. (https://stackoverflow.com/a/37673211/4766178)
        # _cells does not exist for the ReadOnlyWorksheet, no way of knowing
        # is there data or not
        # pylint: disable=protected-access
        is_readonly = isinstance(sheet, openpyxl.worksheet._read_only.ReadOnlyWorksheet)
        return (
            False if is_readonly else not sheet._cells
        )  # there's no public API for this

    @property
    def sheetnames(self):
        return list(self._book.sheetnames)

    @property
    def active(self):
        if not self._active:
            self._active = self._book.active.title

        return self._active

    @active.setter
    def active(self, value):
        if isinstance(value, int):
            value = self.sheetnames[value]
        elif value not in self.sheetnames:
            raise ValueError(f"Unknown worksheet: {value}")

        self._book.active = self.sheetnames.index(value)
        self._active = value

    @property
    def extension(self):
        return self._extension

    @property
    def read_only(self):
        return self._read_only

    def _get_sheetname(self, name=None):
        if not self.sheetnames:
            raise ValueError("No worksheets in file")

        if name is None:
            name = self.active
        elif isinstance(name, int):
            name = self.sheetnames[name]

        return name

    def _get_cellname(self, row, column):
        row = int(row)
        try:
            column = int(column)
            column = get_column_letter(column)
        except ValueError:
            pass
        return "%s%s" % (column, row)

    def _to_index(self, value):
        value = int(value) if value is not None else 1
        if value < 1:
            raise ValueError("Invalid row index")
        return value

    def create(self):
        self._book = openpyxl.Workbook()
        self._extension = None

    def open(self, path=None, read_only=False, write_only=False, data_only=False):
        self._read_only = read_only
        path = path or self.path
        if not path:
            raise ValueError("No path defined for workbook")

        try:
            extension = pathlib.Path(path).suffix
        except TypeError:
            extension = None

        options = {"filename": path, "data_only": data_only}

        # Only set mode arguments if truthy, otherwise openpyxl complains
        if read_only and write_only:
            raise ValueError("Unable to use both write_only and read_only mode")
        elif read_only:
            options["read_only"] = True
        elif write_only:
            options["write_only"] = True

        if extension in (".xlsm", ".xltm"):
            options["keep_vba"] = True

        self._book = openpyxl.load_workbook(**options)
        self._extension = extension

    def close(self):
        self._book.close()
        self._book = None
        self._extension = None
        self._active = None

    def validate_content(self):
        self._validate_content(self._book.properties)

    def save(self, path=None):
        path = path or self.path
        if not path:
            raise ValueError("No path defined for workbook")

        self._book.save(filename=path)

    def create_worksheet(self, name):
        self._book.create_sheet(title=name)
        self.active = name

    def read_worksheet(self, name=None, header=False, start=None) -> List[dict]:
        name = self._get_sheetname(name)
        sheet = self._book[name]
        start = self._to_index(start)

        if start > sheet.max_row or self.is_sheet_empty(sheet):
            return []

        if header:
            columns = [cell.value for cell in sheet[start]]
            start += 1
        else:
            columns = [get_column_letter(i + 1) for i in range(sheet.max_column)]

        columns = [str(value) if value is not None else value for value in columns]
        columns = ensure_unique(columns)

        data = []
        for cells in sheet.iter_rows(min_row=start):
            row = {}
            for c, acell in enumerate(cells):
                column = columns[c]
                if column is not None:
                    row[column] = acell.value
            data.append(row)

        self.active = name
        return data

    def append_worksheet(
        self,
        name=None,
        content=None,
        header=False,
        start=None,
        formatting_as_empty=False,
    ):
        content = Table(content)
        if not content:
            return

        sheet_name = self._get_sheetname(name)
        sheet = self._book[sheet_name]
        start = self._to_index(start)
        is_empty = self.is_sheet_empty(sheet)

        if header and not is_empty:
            columns = [cell.value for cell in sheet[start]]
        else:
            columns = content.columns

        if header and is_empty:
            sheet.append(columns)

        if formatting_as_empty:
            self._append_on_first_empty_based_on_values(content, columns, sheet)
        else:
            self._default_append_rows(content, columns, sheet)

        self.active = sheet_name

    def _append_on_first_empty_based_on_values(self, content, columns, sheet):
        first_empty_row: Optional[int] = None
        for row_num in range(sheet.max_row, 0, -1):
            if all(cell.value is None for cell in sheet[row_num]):
                first_empty_row = row_num
            else:
                break
        first_empty_row: int = first_empty_row or sheet.max_row + 1
        for row_idx, row in enumerate(content):
            values = self._row_to_values(row, columns)
            for cell_idx, acell in enumerate(sheet[first_empty_row + row_idx]):
                try:
                    acell.value = values[cell_idx]
                except IndexError:
                    pass

    def _default_append_rows(self, content, columns, sheet):
        for row in content:
            values = self._row_to_values(row, columns)
            sheet.append(values)

    def _row_to_values(self, row, columns):
        values = [""] * len(columns)
        for column, value in row.items():
            try:
                index = columns.index(column)
                values[index] = value
            except ValueError:
                pass
        return values

    def remove_worksheet(self, name=None):
        name = self._get_sheetname(name)
        others = [sheet for sheet in self.sheetnames if sheet != name]

        if not others:
            raise ValueError("Workbook must have at least one other worksheet")

        if name == self.active:
            self.active = others[0]

        sheet = self._book[name]
        self._book.remove(sheet)

    def rename_worksheet(self, title, name=None):
        title = str(title)
        name = self._get_sheetname(name)
        sheet = self._book[name]

        sheet.title = title
        self.active = title

    def find_empty_row(self, name=None):
        name = self._get_sheetname(name)
        sheet = self._book[name]

        for idx in reversed(range(sheet.max_row)):
            idx += 1  # Convert to 1-based indexing
            if any(value for value in sheet[idx]):
                return idx + 1  # Return first empty row

        return 1

    def get_cell_value(self, row, column, name=None):
        name = self._get_sheetname(name)
        sheet = self._book[name]
        cell = self._get_cellname(row, column)

        return sheet[cell].value

    def set_cell_value(self, row, column, value, name=None):
        name = self._get_sheetname(name)
        sheet = self._book[name]
        cell = self._get_cellname(row, column)

        sheet[cell].value = value

    def set_cell_format(self, row, column, fmt, name=None):
        name = self._get_sheetname(name)
        sheet = self._book[name]
        cell = self._get_cellname(row, column)

        sheet[cell].number_format = str(fmt)

    def insert_image(self, row, column, image, name=None):
        name = self._get_sheetname(name)
        sheet = self._book[name]
        cell = self._get_cellname(row, column)

        # For compatibility with openpyxl
        stream = BytesIO()
        image.save(stream, format=image.format)
        image.fp = stream

        img = openpyxl.drawing.image.Image(image)
        img.anchor = cell
        sheet.add_image(img)


class XlsWorkbook(BaseWorkbook):
    """Container for manipulating legacy Excel files (.xls)"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._images = []

    @staticmethod
    def is_sheet_empty(sheet):
        return not any([sheet.ncols, sheet.nrows])

    @property
    def sheetnames(self):
        return [sheet.name for sheet in self._book.sheets()]

    @property
    def active(self):
        if not self._active:
            for sheet in self._book.sheets():
                if sheet.sheet_visible:
                    self._active = sheet.name
                    break
            else:
                self._active = self.sheetnames[0]

        return self._active

    @active.setter
    def active(self, value):
        if isinstance(value, int):
            value = self.sheetnames[value]
        elif value not in self.sheetnames:
            raise ValueError(f"Unknown worksheet: {value}")

        for sheet in self._book.sheets():
            match = int(sheet.name == value)
            sheet.sheet_selected = match
            sheet.sheet_visible = match

        self._active = value

    @property
    def extension(self):
        return self._extension

    def _get_sheetname(self, name):
        if self._book.nsheets == 0:
            raise ValueError("No worksheets in file")

        if name is None:
            name = self.active
        elif isinstance(name, int):
            name = self.sheetnames[name]
        elif name not in self.sheetnames:
            raise ValueError(f"Unknown worksheet: {name}")

        return name

    def _get_cell(self, row, column):
        row = int(row)
        try:
            column = int(column)
        except ValueError:
            column = get_column_index(column)
        return row - 1, column - 1

    def _to_index(self, value):
        value = (int(value) - 1) if value is not None else 0
        if value < 0:
            raise ValueError("Invalid row index")
        return value

    def create(self, sheet="Sheet"):
        fd = BytesIO()
        try:
            book = xlwt.Workbook()
            book.add_sheet(sheet)
            book.save(fd)
            fd.seek(0)
            self.open(fd)
        finally:
            fd.close()

        self._extension = None

    def open(self, path=None, read_only=False, write_only=False, data_only=False):
        path = path or self.path
        if not path:
            raise ValueError("No path defined for workbook")

        try:
            extension = pathlib.Path(path).suffix
        except TypeError:
            extension = None

        options = {"on_demand": True, "formatting_info": True}

        if read_only or write_only or data_only:
            self.logger.info(
                "Modes read_only/write_only/data_only not supported with .xls"
            )

        if hasattr(path, "read"):
            options["file_contents"] = path.read()
        else:
            options["filename"] = path

        self._book = xlrd.open_workbook(**options)
        self._extension = extension
        self._images = []

    def close(self):
        self._book.release_resources()
        self._book = None
        self._extension = None
        self._active = None
        self._images = []

    @contextmanager
    def _book_write(self):
        book = xlutils_copy(self._book)
        yield book

        fd = BytesIO()
        try:
            book.save(fd)
            fd.seek(0)
            self.close()
            self.open(fd)
        finally:
            fd.close()

    def validate_content(self):
        self._validate_content(self._book)

    def save(self, path=None):
        path = path or self.path
        if not path:
            raise ValueError("No path defined for workbook")

        book = xlutils_copy(self._book)
        self._insert_images(book)
        book.save(path)

    def create_worksheet(self, name):
        with self._book_write() as book:
            book.add_sheet(name)

        self.active = name

    def read_worksheet(self, name=None, header=False, start=None) -> List[dict]:
        name = self._get_sheetname(name)
        sheet = self._book.sheet_by_name(name)
        start = self._to_index(start)

        if start >= sheet.nrows:
            return []

        if header:
            columns = [self._parse_type(cell) for cell in sheet.row(start)]
            start += 1
        else:
            columns = [get_column_letter(i + 1) for i in range(sheet.ncols)]

        columns = [value if value != "" else None for value in columns]
        columns = [str(value) if value is not None else value for value in columns]
        columns = ensure_unique(columns)

        data = []
        for r in range(start, sheet.nrows):
            row = {}
            for c in range(sheet.ncols):
                column = columns[c]
                if column is not None:
                    cell = sheet.cell(r, c)
                    row[column] = self._parse_type(cell)
            data.append(row)

        self.active = name
        return data

    def _parse_type(self, cell):
        value = cell.value

        if cell.ctype == xlrd.XL_CELL_DATE:
            value = xlrd.xldate_as_datetime(value, self._book.datemode)
        elif cell.ctype == xlrd.XL_CELL_BOOLEAN:
            value = bool(value)
        elif cell.ctype == xlrd.XL_CELL_ERROR:
            value = xlrd.biffh.error_text_from_code.get(value, "#ERROR")
        elif cell.ctype == xlrd.XL_CELL_NUMBER and value.is_integer():
            value = int(value)

        return value

    def append_worksheet(
        self,
        name=None,
        content=None,
        header=False,
        start=None,
        formatting_as_empty=False,
    ):
        content = Table(content)
        if not content:
            return

        name = self._get_sheetname(name)
        sheet_read = self._book.sheet_by_name(name)
        start = self._to_index(start)
        is_empty = self.is_sheet_empty(sheet_read)

        if header and not is_empty:
            columns = [cell.value for cell in sheet_read.row(start)]
        else:
            columns = content.columns

        with self._book_write() as book:
            sheet_write = book.get_sheet(name)
            # TODO. target worksheet cell formatting is overwritten.
            # It would be preferable to preserve formatting in the
            # target worksheet.
            if formatting_as_empty:
                start_row = self._return_first_empty_row(sheet_read)
            else:
                start_row = sheet_read.nrows
            if header and is_empty:
                for column, value in enumerate(columns):
                    sheet_write.write(0, column, value)
                start_row += 1

            for r, row in enumerate(content, start_row):
                for column, value in row.items():
                    sheet_write.write(r, columns.index(column), value)

        self.active = name

    def _return_first_empty_row(self, sheet):
        first_empty_row: Optional[int] = None
        for row_num in range(sheet.nrows - 1, 0, -1):
            if all(cell.value == "" for cell in sheet[row_num]):
                first_empty_row = row_num
            else:
                break
        first_empty_row: int = first_empty_row or sheet.nrows
        return first_empty_row

    def remove_worksheet(self, name=None):
        name = self._get_sheetname(name)
        others = [sheet for sheet in self.sheetnames if sheet != name]

        if not others:
            raise ValueError("Workbook must have at least one other worksheet")

        if name == self.active:
            self.active = others[0]

        with self._book_write() as book:
            # This is pretty ugly, but there seems to be no other way to
            # remove sheets from the xlwt.Workbook instance
            # pylint: disable=protected-access
            book._Workbook__worksheets = [
                sheet for sheet in book._Workbook__worksheets if sheet.name != name
            ]
            book._Workbook__active_sheet = next(
                idx
                for idx, sheet in enumerate(book._Workbook__worksheets)
                if sheet.name == self.active
            )

    def rename_worksheet(self, title, name=None):
        title = str(title)
        name = self._get_sheetname(name)

        with self._book_write() as book:
            sheet = book.get_sheet(name)
            sheet.name = title

        self.active = title

    def find_empty_row(self, name=None):
        name = self._get_sheetname(name)
        sheet = self._book.sheet_by_name(name)

        for row in reversed(range(sheet.nrows)):
            if any(cell.value for cell in sheet.row(row)):
                # Convert to 1-based indexing and
                # return first empty row
                return row + 2

        return 1

    def get_cell_value(self, row, column, name=None):
        name = self._get_sheetname(name)
        sheet = self._book.sheet_by_name(name)
        row, column = self._get_cell(row, column)

        return sheet.cell_value(row, column)

    def set_cell_value(self, row, column, value, name=None):
        name = self._get_sheetname(name)
        row, column = self._get_cell(row, column)

        with self._book_write() as book:
            sheet = book.get_sheet(name)
            sheet.write(row, column, value)

    def set_cell_format(self, row, column, fmt, name=None):
        name = self._get_sheetname(name)
        sheet = self._book.sheet_by_name(name)
        row, column = self._get_cell(row, column)

        value = sheet.cell_value(row, column)
        style = xlwt.XFStyle()
        style.num_format_str = str(fmt)

        with self._book_write() as book:
            sheet = book.get_sheet(name)
            sheet.write(row, column, value, style)

    def insert_image(self, row, column, image, name=None):
        name = self._get_sheetname(name)
        row, column = self._get_cell(row, column)
        self._images.append((name, row, column, image))

    def _insert_images(self, book):
        for name, row, column, image in self._images:
            stream = BytesIO()
            image.save(stream, format="BMP")
            bitmap = stream.getvalue()

            sheet = book.get_sheet(name)
            sheet.insert_bitmap_data(bitmap, row, column)



from pathlib import Path
from PIL import Image

_RESOURCE_DIR = Path(__file__).parent.parent / "resources"
_IMAGE_PATH = _RESOURCE_DIR / "images" / "source.png"

_MOCK_IMAGE_ATTACHMENT = Image.open(_IMAGE_PATH)
_MOCK_IMAGE_ATTACHMENT.load()


# Exported constants
MOCK_ACCESS_TOKEN = "smartsheetAccessToken{:0>2}"
"""You must use with .format(<int>)"""
MOCK_SHEET_JSON = {
    "id": 7340596407887748,
    "name": "orders",
    "version": 10,
    "totalRowCount": 9,
    "accessLevel": "OWNER",
    "effectiveAttachmentOptions": [
        "GOOGLE_DRIVE",
        "DROPBOX",
        "BOX_COM",
        "EVERNOTE",
        "ONEDRIVE",
        "LINK",
        "FILE",
        "EGNYTE",
    ],
    "ganttEnabled": False,
    "dependenciesEnabled": False,
    "resourceManagementEnabled": False,
    "resourceManagementType": "NONE",
    "cellImageUploadEnabled": True,
    "userSettings": {"criticalPathEnabled": False, "displaySummaryTasks": True},
    "permalink": "https://app.smartsheet.com/sheets/g4J274W3V7F549FvxPVrrWV3fJGCC9xGqjq9Gqm1",
    "createdAt": "2023-02-17T19:00:58Z",
    "modifiedAt": "2023-02-22T22:46:45Z",
    "isMultiPicklistEnabled": True,
    "columns": [
        {
            "id": 3471889768703876,
            "version": 0,
            "index": 0,
            "title": "Name",
            "type": "TEXT_NUMBER",
            "primary": True,
            "validation": False,
            "width": 144,
        },
        {
            "id": 7975489396074372,
            "version": 0,
            "index": 1,
            "title": "Item",
            "type": "TEXT_NUMBER",
            "validation": False,
            "width": 294,
        },
        {
            "id": 657140001597316,
            "version": 0,
            "index": 2,
            "title": "Zip",
            "type": "TEXT_NUMBER",
            "validation": False,
            "width": 51,
        },
    ],
    "rows": [
        {
            "id": 3218304495249284,
            "rowNumber": 1,
            "expanded": True,
            "createdAt": "2023-02-17T19:00:58Z",
            "modifiedAt": "2023-02-21T21:18:43Z",
            "cells": [
                {
                    "columnId": 3471889768703876,
                    "value": "Sol Heaton",
                    "displayValue": "Sol Heaton",
                },
                {
                    "columnId": 7975489396074372,
                    "value": "Sauce Labs Bolt T-Shirt",
                    "displayValue": "Sauce Labs Bolt T-Shirt",
                },
                {
                    "columnId": 657140001597316,
                    "value": 3695.0,
                    "displayValue": "3695",
                },
            ],
        },
        {
            "id": 7721904122619780,
            "rowNumber": 2,
            "siblingId": 3218304495249284,
            "expanded": True,
            "createdAt": "2023-02-17T19:00:58Z",
            "modifiedAt": "2023-02-17T19:00:58Z",
            "cells": [
                {
                    "columnId": 3471889768703876,
                    "value": "Gregg Arroyo",
                    "displayValue": "Gregg Arroyo",
                },
                {
                    "columnId": 7975489396074372,
                    "value": "Sauce Labs Onesie",
                    "displayValue": "Sauce Labs Onesie",
                },
                {
                    "columnId": 657140001597316,
                    "value": 4418.0,
                    "displayValue": "4418",
                },
            ],
        },
        {
            "id": 2092404588406660,
            "rowNumber": 3,
            "siblingId": 7721904122619780,
            "expanded": True,
            "createdAt": "2023-02-17T19:00:58Z",
            "modifiedAt": "2023-02-17T19:00:58Z",
            "cells": [
                {
                    "columnId": 3471889768703876,
                    "value": "Zoya Roche",
                    "displayValue": "Zoya Roche",
                },
                {
                    "columnId": 7975489396074372,
                    "value": "Sauce Labs Bolt T-Shirt",
                    "displayValue": "Sauce Labs Bolt T-Shirt",
                },
                {
                    "columnId": 657140001597316,
                    "value": 3013.0,
                    "displayValue": "3013",
                },
            ],
        },
        {
            "id": 6596004215777156,
            "rowNumber": 4,
            "siblingId": 2092404588406660,
            "expanded": True,
            "createdAt": "2023-02-17T19:00:58Z",
            "modifiedAt": "2023-02-17T19:00:58Z",
            "cells": [
                {
                    "columnId": 3471889768703876,
                    "value": "Gregg Arroyo",
                    "displayValue": "Gregg Arroyo",
                },
                {
                    "columnId": 7975489396074372,
                    "value": "Sauce Labs Bolt T-Shirt",
                    "displayValue": "Sauce Labs Bolt T-Shirt",
                },
                {
                    "columnId": 657140001597316,
                    "value": 4418.0,
                    "displayValue": "4418",
                },
            ],
        },
        {
            "id": 4344204402091908,
            "rowNumber": 5,
            "siblingId": 6596004215777156,
            "expanded": True,
            "createdAt": "2023-02-17T19:00:58Z",
            "modifiedAt": "2023-02-17T19:00:58Z",
            "cells": [
                {
                    "columnId": 3471889768703876,
                    "value": "Camden Martin",
                    "displayValue": "Camden Martin",
                },
                {
                    "columnId": 7975489396074372,
                    "value": "Sauce Labs Bolt T-Shirt",
                    "displayValue": "Sauce Labs Bolt T-Shirt",
                },
                {
                    "columnId": 657140001597316,
                    "value": 1196.0,
                    "displayValue": "1196",
                },
            ],
        },
        {
            "id": 8847804029462404,
            "rowNumber": 6,
            "siblingId": 4344204402091908,
            "expanded": True,
            "createdAt": "2023-02-17T19:00:58Z",
            "modifiedAt": "2023-02-17T19:00:58Z",
            "cells": [
                {
                    "columnId": 3471889768703876,
                    "value": "Zoya Roche",
                    "displayValue": "Zoya Roche",
                },
                {
                    "columnId": 7975489396074372,
                    "value": "Sauce Labs Fleece Jacket",
                    "displayValue": "Sauce Labs Fleece Jacket",
                },
                {
                    "columnId": 657140001597316,
                    "value": 3013.0,
                    "displayValue": "3013",
                },
            ],
        },
        {
            "id": 262817239787396,
            "rowNumber": 7,
            "siblingId": 8847804029462404,
            "expanded": True,
            "createdAt": "2023-02-17T19:00:58Z",
            "modifiedAt": "2023-02-17T19:00:58Z",
            "cells": [
                {
                    "columnId": 3471889768703876,
                    "value": "Zoya Roche",
                    "displayValue": "Zoya Roche",
                },
                {
                    "columnId": 7975489396074372,
                    "value": "Sauce Labs Onesie",
                    "displayValue": "Sauce Labs Onesie",
                },
                {
                    "columnId": 657140001597316,
                    "value": 3013.0,
                    "displayValue": "3013",
                },
            ],
        },
        {
            "id": 4766416867157892,
            "rowNumber": 8,
            "siblingId": 262817239787396,
            "expanded": True,
            "createdAt": "2023-02-17T19:00:58Z",
            "modifiedAt": "2023-02-17T19:00:58Z",
            "cells": [
                {
                    "columnId": 3471889768703876,
                    "value": "Sol Heaton",
                    "displayValue": "Sol Heaton",
                },
                {
                    "columnId": 7975489396074372,
                    "value": "Sauce Labs Fleece Jacket",
                    "displayValue": "Sauce Labs Fleece Jacket",
                },
                {
                    "columnId": 657140001597316,
                    "value": 3695.0,
                    "displayValue": "3695",
                },
            ],
        },
        {
            "id": 2514617053472644,
            "rowNumber": 9,
            "siblingId": 4766416867157892,
            "expanded": True,
            "createdAt": "2023-02-17T19:00:58Z",
            "modifiedAt": "2023-02-17T19:00:58Z",
            "cells": [
                {
                    "columnId": 3471889768703876,
                    "value": "Sol Heaton",
                    "displayValue": "Sol Heaton",
                },
                {
                    "columnId": 7975489396074372,
                    "value": "Sauce Labs Onesee",
                    "displayValue": "Sauce Labs Onesee",
                },
                {
                    "columnId": 657140001597316,
                    "value": 3695.0,
                    "displayValue": "3695",
                },
            ],
        },
    ],
}
"""A mocked sheet response"""
MOCK_IMAGE = _MOCK_IMAGE_ATTACHMENT.tobytes()
"""An image in bytes"""
MOCK_IMAGE_ATTACHMENT = {
    "expected_filename": "source.png",
    "expected_content": MOCK_IMAGE,
    "mime_type": "image/png",
}
"""A mocked image attachment"""
MOCK_TEXT_ATTACHMENT = {
    "expected_filename": "note01.txt",
    "expected_content": "Hello World! I'm note number 01!",
    "mime_type": "text/plain",
}
"""A mocked text attachment, use with `.format(<int>)`."""
ATTACHMENT_NODE = {
    "id": None,
    "name": None,
    "attachmentType": "FILE",
    "mimeType": None,
    "sizeInKb": 1,
    "createdAt": "2023-02-22T22:47:26Z",
    "createdBy": {"email": "markmonkey@robocorp.com"},
}
"""You should set the `id`, `name`, and `mimeType` attributes of this dict."""
_attach_one = ATTACHMENT_NODE.copy()
_attach_one["id"] = 123
_attach_one["name"] = "source.png"
_attach_one["mimeType"] = "image/png"
_attach_two = ATTACHMENT_NODE.copy()
_attach_two["id"] = 456
_attach_two["name"] = "note01.txt"
_attach_two["mimeType"] = "text/plain"
MOCKED_ATTACHMENT_LIST = [_attach_one, _attach_two]
"""A mocked attachment list"""
MOCKED_ROW_ATTACHMENTS = {
    "pageNumber": 1,
    "totalPages": 1,
    "totalCount": 2,
    "data": MOCKED_ATTACHMENT_LIST,
}
"""A mocked response for row attachments"""
MOCKED_ROW_NO_ATTACHMENTS = {
    "pageNumber": 0,
    "pageSize": 100,
    "totalPages": 0,
    "totalCount": 0,
    "data": [],
}
"""A mocked response for a row with no attachments"""
MOCKED_ROW = {
    "id": 3218304495249284,
    "sheetId": 7340596407887748,
    "rowNumber": 1,
    "version": 10,
    "permalink": "https://app.smartsheet.com/sheets/g4J274W3V7F549FvxPVrrWV3fJGCC9xGqjq9Gqm1?rowId=3218304495249284",
    "filteredOut": False,
    "expanded": True,
    "accessLevel": "OWNER",
    "createdAt": "2023-02-17T19:00:58Z",
    "createdBy": {"email": "markmonkey@robocorp.com"},
    "modifiedAt": "2023-02-21T21:18:43Z",
    "modifiedBy": {"email": "markmonkey@robocorp.com"},
    "cells": [
        {
            "columnId": 3471889768703876,
            "columnType": "TEXT_NUMBER",
            "value": "Sol Heaton",
            "objectValue": "Sol Heaton",
            "displayValue": "Sol Heaton",
            "format": ",,,,,,,3,1,2,,,,,,,",
        },
        {
            "columnId": 7975489396074372,
            "columnType": "TEXT_NUMBER",
            "value": "Sauce Labs Bolt T-Shirt",
            "objectValue": "Sauce Labs Bolt T-Shirt",
            "displayValue": "Sauce Labs Bolt T-Shirt",
            "format": ",,,,,,,3,1,2,,,,,,,",
        },
        {
            "columnId": 657140001597316,
            "columnType": "TEXT_NUMBER",
            "value": 3695.0,
            "objectValue": 3695.0,
            "displayValue": "3695",
            "format": ",,,,,,,3,1,2,,,,,,,",
        },
    ],
    "attachments": [
        {
            "id": 1037987386288004,
            "name": "test.zip",
            "attachmentType": "FILE",
            "mimeType": "application/zip",
            "sizeInKb": 1,
            "createdAt": "2023-02-22T22:47:26Z",
            "createdBy": {"email": "markmonkey@robocorp.com"},
        },
        {
            "id": 19892232316804,
            "name": "bitmap-Monkey learning processes.png",
            "attachmentType": "FILE",
            "mimeType": "image/png",
            "sizeInKb": 302,
            "createdAt": "2023-02-22T22:25:27Z",
            "createdBy": {"email": "markmonkey@robocorp.com"},
        },
    ],
}
NEW_ROW_1 = [
    {"title": "Name", "value": "Mark Monkey"},
    {"title": "Item", "value": "Sauce Labs Onesie"},
    {"title": "Zip", "value": "California"},
]
"""A mocked row to be added to a sheet"""
NEW_ROW_2 = [
    {"title": "Name", "value": "Mark Monkey"},
    {"title": "Item", "value": "Sauce Labs Fleece Jacket"},
    {"title": "Zip", "value": 90210},
]
NEW_ROWS = [NEW_ROW_1, NEW_ROW_2]
"""A list of mocked rows to be added to a sheet"""
MOCKED_ADD_ROW_SUCCESS = {
    "message": "SUCCESS",
    "resultCode": 0,
    "result": [
        {
            "id": 7859947312768900,
            "sheetId": 7340596407887748,
            "rowNumber": 10,
            "siblingId": 2514617053472644,
            "expanded": True,
            "createdAt": "2023-03-06T20:50:07Z",
            "modifiedAt": "2023-03-06T20:50:07Z",
            "cells": [
                {
                    "columnId": 3471889768703876,
                    "value": "Mark Monkey",
                    "displayValue": "Mark Monkey",
                },
                {
                    "columnId": 7975489396074372,
                    "value": "Sauce Labs Onesie",
                    "displayValue": "Sauce Labs Onesie",
                },
                {
                    "columnId": 657140001597316,
                    "value": "California",
                    "displayValue": "California",
                },
            ],
        },
        {
            "id": 2230447778555780,
            "sheetId": 7340596407887748,
            "rowNumber": 11,
            "siblingId": 7859947312768900,
            "expanded": True,
            "createdAt": "2023-03-06T20:50:07Z",
            "modifiedAt": "2023-03-06T20:50:07Z",
            "cells": [
                {
                    "columnId": 3471889768703876,
                    "value": "Mark Monkey",
                    "displayValue": "Mark Monkey",
                },
                {
                    "columnId": 7975489396074372,
                    "value": "Sauce Labs Fleece Jacket",
                    "displayValue": "Sauce Labs Fleece Jacket",
                },
                {"columnId": 657140001597316, "value": 90210, "displayValue": "90210"},
            ],
        },
    ],
    "version": 11,
}
"""A mocked response for a successful row add"""
UPDATED_ROW_1 = [
    {"title": "Name", "value": "Mark Monkey Jr"},
    {"title": "Item", "value": "Sauce Labs Onesie"},
    {"title": "Zip", "value": 90210},
]
"""A mocked row to be updated in a sheet"""
UPDATED_ROW_2 = [
    {"title": "Name", "value": "Mark Monkey Jr"},
    {"title": "Item", "value": "Sauce Labs Fleece Jacket"},
    {"title": "Zip", "value": 90210},
]
_row_1_with_id = UPDATED_ROW_1.copy()
_row_1_with_id.insert(0, {"rowId": 7859947312768900})
_row_2_with_id = UPDATED_ROW_2.copy()
_row_2_with_id.insert(0, {"rowId": 2230447778555780})
UPDATED_ROWS = [_row_1_with_id, _row_2_with_id]
"""A set of mocked rows to be updated in a sheet"""
NAME_COLUMN_ID = 3471889768703876
ITEM_COLUMN_ID = 7975489396074372
ZIP_COLUMN_ID = 657140001597316
MOCKED_UPDATE_ROW_SUCCESS = {
    "message": "SUCCESS",
    "resultCode": 0,
    "result": [
        {
            "id": 7859947312768900,
            "rowNumber": 1,
            "expanded": True,
            "createdAt": "2023-03-06T20:50:07Z",
            "modifiedAt": "2023-03-06T20:51:53Z",
            "cells": [
                {
                    "columnId": 3471889768703876,
                    "value": "Mark Monkey Jr",
                    "displayValue": "Mark Monkey Jr",
                },
                {
                    "columnId": 7975489396074372,
                    "value": "Sauce Labs Onesie",
                    "displayValue": "Sauce Labs Onesie",
                },
                {"columnId": 657140001597316, "value": 90210, "displayValue": "90210"},
            ],
        },
        {
            "id": 2230447778555780,
            "rowNumber": 2,
            "siblingId": 7859947312768900,
            "expanded": True,
            "createdAt": "2023-03-06T20:50:07Z",
            "modifiedAt": "2023-03-06T20:51:53Z",
            "cells": [
                {
                    "columnId": 3471889768703876,
                    "value": "Mark Monkey Jr",
                    "displayValue": "Mark Monkey Jr",
                },
                {
                    "columnId": 7975489396074372,
                    "value": "Sauce Labs Fleece Jacket",
                    "displayValue": "Sauce Labs Fleece Jacket",
                },
                {"columnId": 657140001597316, "value": 90210, "displayValue": "90210"},
            ],
        },
    ],
    "version": 12,
}
"""A mocked response for a successful row update"""
NEW_SHEET_COLUMNS = [
    {"title": "Order Number", "type": "TEXT_NUMBER", "primary": True},
    {"title": "Customer Name", "type": "TEXT_NUMBER"},
    {"title": "Order Date", "type": "DATE"},
    {
        "title": "Items",
        "type": "MULTI_PICKLIST",
        "options": [
            "Sauce Labs Backpack",
            "Sauce Labs Bike Light",
            "Sauce Labs Bolt T-Shirt",
            "Sauce Labs Fleece Jacket",
            "Sauce Labs Onesie",
            "Test.allTheThings() T-Shirt (Red)",
        ],
    },
]
"""A mocked set of columns to be added to a new sheet"""
NEW_SHEET_NAME = "Test Sheet"
"""A mocked name for a new sheet"""
NEW_SHEET = {
    "name": NEW_SHEET_NAME,
    "columns": NEW_SHEET_COLUMNS,
}
"""A mocked new sheet object."""
MOCKED_CREATE_SHEET_SUCCESS = {
    "message": "SUCCESS",
    "resultCode": 0,
    "result": {
        "id": 842641693796228,
        "name": "Test Sheet",
        "accessLevel": "OWNER",
        "permalink": "https://app.smartsheet.com/sheets/v73m5PQq7g4W7cGmVc7CHPR5v74JG6wQvMFj5331",
        "columns": [
            {
                "id": 1502033234159492,
                "version": 0,
                "index": 0,
                "title": "Order Number",
                "type": "TEXT_NUMBER",
                "primary": True,
                "validation": False,
                "width": 150,
            },
            {
                "id": 6005632861529988,
                "version": 0,
                "index": 1,
                "title": "Customer Name",
                "type": "TEXT_NUMBER",
                "validation": False,
                "width": 150,
            },
            {
                "id": 3753833047844740,
                "version": 0,
                "index": 2,
                "title": "Order Date",
                "type": "DATE",
                "validation": False,
                "width": 150,
            },
            {
                "id": 8257432675215236,
                "version": 2,
                "index": 3,
                "title": "Items",
                "type": "MULTI_PICKLIST",
                "options": [
                    "Sauce Labs Backpack",
                    "Sauce Labs Bike Light",
                    "Sauce Labs Bolt T-Shirt",
                    "Sauce Labs Fleece Jacket",
                    "Sauce Labs Onesie",
                    "Test.allTheThings() T-Shirt (Red)",
                ],
                "validation": False,
                "width": 150,
            },
        ],
    },
}
"""A mocked response for a successful sheet creation"""
MOCKED_SEARCH_RESULTS = {
    "results": [
        {
            "text": "Sauce Labs Onesie",
            "objectType": "row",
            "objectId": 7859947312768900,
            "parentObjectType": "sheet",
            "parentObjectId": 7340596407887748,
            "parentObjectName": "orders",
            "contextData": ["Mark Monkey Jr"],
            "parentObjectFavorite": True,
        },
        {
            "text": "Sauce Labs Onesie",
            "objectType": "row",
            "objectId": 7721904122619780,
            "parentObjectType": "sheet",
            "parentObjectId": 7340596407887748,
            "parentObjectName": "orders",
            "contextData": ["Gregg Arroyo"],
            "parentObjectFavorite": True,
        },
        {
            "text": "Sauce Labs Onesie",
            "objectType": "row",
            "objectId": 262817239787396,
            "parentObjectType": "sheet",
            "parentObjectId": 7340596407887748,
            "parentObjectName": "orders",
            "contextData": ["Zoya Roche"],
            "parentObjectFavorite": True,
        },
        {
            "text": "Sauce Labs Onesee",
            "objectType": "row",
            "objectId": 2514617053472644,
            "parentObjectType": "sheet",
            "parentObjectId": 7340596407887748,
            "parentObjectName": "orders",
            "contextData": ["Sol Heaton"],
            "parentObjectFavorite": True,
        },
    ],
    "totalCount": 4,
}



"""Common utilities shared by any e-mail related library."""

import os
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Optional, Union

import jwt

from RPA.MFA import MFA


lib_mfa = MFA()

OAuthProviderType = Union["OAuthProvider", str]


class NoRecipientsError(ValueError):
    """Raised when email to be sent does not have any recipients, cc or bcc addresses."""  # noqa: E501


class OAuthProvider(Enum):
    """OAuth2 tested providers."""

    GOOGLE = "google"
    MICROSOFT = "microsoft"


@dataclass
class OAuthConfig:
    """OAuth2 Authorization Code Flow provider settings."""

    auth_url: str
    redirect_uri: str
    scope: str
    token_url: str


OAUTH_PROVIDERS = {
    OAuthProvider.GOOGLE: OAuthConfig(
        auth_url="https://accounts.google.com/o/oauth2/auth",
        redirect_uri="https://developers.google.com/oauthplayground",
        scope="https://mail.google.com",
        token_url="https://accounts.google.com/o/oauth2/token",
    ),
    OAuthProvider.MICROSOFT: OAuthConfig(
        auth_url="https://login.microsoftonline.com/{tenant}/oauth2/v2.0/authorize",
        redirect_uri="https://login.microsoftonline.com/common/oauth2/nativeclient",
        scope="offline_access https://outlook.office365.com/.default",
        token_url="https://login.microsoftonline.com/{tenant}/oauth2/v2.0/token",
    ),
}


class OAuthMixin:
    """Common keywords for the Email libraries, enabling OAuth2 support."""

    TO_PROTECT = ["get_oauth_token", "refresh_oauth_token"]

    def __init__(self, provider: OAuthProviderType, tenant: Optional[str]):
        self._oauth_provider = OAUTH_PROVIDERS[OAuthProvider(provider)]
        if tenant:
            for url_attr in ("auth_url", "redirect_uri", "token_url"):
                formatted = getattr(self._oauth_provider, url_attr).format(
                    tenant=tenant
                )
                setattr(self._oauth_provider, url_attr, formatted)
        # NOTE(cmin764):
        #  https://github.com/requests/requests-oauthlib/issues/387#issuecomment-1325131664
        os.environ["OAUTHLIB_RELAX_TOKEN_SCOPE"] = "1"

    def generate_oauth_url(self, client_id: str) -> str:
        """Generates an authorization URL which must be opened by the user to start the
        OAuth2 flow and obtain an authorization code as response.

        :param client_id: Client app ID. (generated by the provider)
        :returns: Authorization URL string not containing any sensitive info in it.

        **Example: Robot Framework**

        .. code-block:: robotframework

            *** Tasks ***
            Start OAuth Flow
                ${auth_url} =    Generate OAuth URL
                ...     client_id=810482312368-19htmcgcj*******googleusercontent.com
                Log     Start OAuth2 flow: ${auth_url}
        """
        return lib_mfa.generate_oauth_url(
            self._oauth_provider.auth_url,
            client_id=client_id,
            redirect_uri=self._oauth_provider.redirect_uri,
            scope=self._oauth_provider.scope,
            access_type="offline",
            prompt="consent",
        )

    def _sync_token_metadata(self, token: dict):
        # Ensures that the issued expiry time in the access token is in sync with the
        #  one we have stored in the token dictionary. So `exchangelib` will know for
        #  sure when to refresh a potentially expired token.
        try:
            access_token = jwt.decode(
                token["access_token"], options={"verify_signature": False}
            )
        except jwt.exceptions.DecodeError:
            # Logger object declared in every derived class.
            self.logger.debug("Couldn't decode access token or token isn't valid JWT")
        else:
            token["expires_at"] = access_token["exp"]

    def get_oauth_token(self, client_secret: str, response_url: str) -> dict:
        """Exchanges the code obtained previously with ``Generate OAuth URL`` for a
        token.

        :param client_secret: Client app secret. (generated by the provider)
        :param response_url: The final URL containing the authorization `code` found in
            the address bar after authenticating and authorizing the Client app
            through the authorization URL.
        :returns: A dictionary containing the access & refresh token, plus metadata.

        **Example: Robot Framework**

        .. code-block:: robotframework

            *** Tasks ***
            Finish OAuth Flow
                ${token} =      Get OAuth Token
                ...     client_secret=GOCSPX-******mqZAW89
                ...     response_url=${resp_url}  # redirect of `Generate OAuth URL`
        """
        token = lib_mfa.get_oauth_token(
            self._oauth_provider.token_url,
            client_secret=client_secret,
            response_url=response_url,
            include_client_id=True,
        )
        self._sync_token_metadata(token)
        return token

    def refresh_oauth_token(
        self, client_id: str, client_secret: str, token: dict
    ) -> dict:
        """Refreshes the token as the access one usually expires after 1h and the
        refresh one never expires. (as long as it doesn't get revoked)

        :param client_id: Client app ID. (generated by the provider)
        :param client_secret: Client app secret. (generated by the provider)
        :param token: Full token dictionary previously obtained with
            ``Get OAuth Token``.
        :returns: A token dictionary containing a new access token and updated
            metadata.

        **Example: Robot Framework**

        .. code-block:: robotframework

            *** Tasks ***
            Refresh OAuth Flow
                ${token} =      Refresh OAuth Token
                ...     client_id=810482312368-19htmcgcj*******googleusercontent.com
                ...     client_secret=GOCSPX-******mqZAW89
                ...     token=${token}  # from `Get OAuth Token`
        """
        token = lib_mfa.refresh_oauth_token(
            self._oauth_provider.token_url,
            client_id=client_id,
            client_secret=client_secret,
            refresh_token=token["refresh_token"],
        )
        self._sync_token_metadata(token)
        return token


def counter_duplicate_path(file_path: Path) -> Path:
    """Returns a unique file path by adding a suffixed counter if already exists."""
    if not file_path.exists():
        return file_path  # unique already

    root_dir = file_path.parent
    duplicates = root_dir.glob(f"{file_path.stem}*{file_path.suffix}")
    suffixes = []
    for dup in duplicates:
        parts = dup.stem.rsplit("-", 1)
        if len(parts) == 2 and parts[1].isdigit():
            suffixes.append(int(parts[1]))
    next_suffix = max(suffixes) + 1 if suffixes else 2

    file_path = root_dir / f"{file_path.stem}-{next_suffix}{file_path.suffix}"
    return file_path
